import json
from dataclasses import dataclass
from os import getenv
from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Tuple, Type, Union

from pydantic import BaseModel

from agno.exceptions import ModelProviderError
from agno.models.base import Model
from agno.models.message import Message
from agno.models.metrics import MessageMetrics
from agno.models.response import ModelResponse
from agno.run.agent import RunOutput
from agno.utils.log import log_debug, log_error, log_warning
from agno.utils.tokens import count_schema_tokens

try:
    from boto3 import client as AwsClient
    from boto3.session import Session
    from botocore.exceptions import ClientError
except ImportError:
    raise ImportError("`boto3` not installed. Please install using `pip install boto3`")

try:
    import aioboto3

    AIOBOTO3_AVAILABLE = True
except ImportError:
    aioboto3 = None
    AIOBOTO3_AVAILABLE = False


BEDROCK_SUPPORTED_IMAGE_FORMATS = ["png", "jpeg", "webp", "gif"]
BEDROCK_SUPPORTED_VIDEO_FORMATS = ["mp4", "mov", "mkv", "webm", "flv", "mpeg", "mpg", "wmv", "three_gp"]
BEDROCK_SUPPORTED_FILE_FORMATS = ["pdf", "csv", "doc", "docx", "xls", "xlsx", "html", "txt", "md"]


@dataclass
class AwsBedrock(Model):
    """
    AWS Bedrock model.

    To use this model, you need to either:
    1. Set the following environment variables:
       - AWS_ACCESS_KEY_ID
       - AWS_SECRET_ACCESS_KEY
       - AWS_REGION
    2. Or provide a boto3 Session object

    For async support, you also need aioboto3 installed:
       pip install aioboto3

    Not all Bedrock models support all features. See this documentation for more information: https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference-supported-models-features.html

    Args:
        aws_region (Optional[str]): The AWS region to use.
        aws_access_key_id (Optional[str]): The AWS access key ID to use.
        aws_secret_access_key (Optional[str]): The AWS secret access key to use.
        aws_sso_auth (Optional[str]): Removes the need for an access and secret access key by leveraging the current profile's authentication
        session (Optional[Session]): A boto3 Session object to use for authentication.
    """

    id: str = "mistral.mistral-small-2402-v1:0"
    name: str = "AwsBedrock"
    provider: str = "AwsBedrock"

    aws_sso_auth: Optional[bool] = False
    aws_region: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None
    session: Optional[Session] = None

    # Request parameters
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    stop_sequences: Optional[List[str]] = None
    request_params: Optional[Dict[str, Any]] = None

    client: Optional[AwsClient] = None
    async_client: Optional[Any] = None
    async_session: Optional[Any] = None

    def get_client(self) -> AwsClient:
        """
        Get the Bedrock client.

        Returns:
            AwsClient: The Bedrock client.
        """
        # When using a boto3 session, always recreate the client so
        # session credentials can be refreshed (IAM roles, EKS, STS).
        if not self.session and self.client is not None:
            return self.client

        # Return directly (not via self.client) so concurrent callers
        # on the same model instance each get their own client.
        if self.session:
            return self.session.client(
                "bedrock-runtime",
                region_name=self.aws_region or self.session.region_name,
            )

        self.aws_access_key_id = self.aws_access_key_id or getenv("AWS_ACCESS_KEY_ID")
        self.aws_secret_access_key = self.aws_secret_access_key or getenv("AWS_SECRET_ACCESS_KEY")
        self.aws_session_token = self.aws_session_token or getenv("AWS_SESSION_TOKEN")
        self.aws_region = self.aws_region or getenv("AWS_REGION")

        if self.aws_sso_auth:
            self.client = AwsClient(service_name="bedrock-runtime", region_name=self.aws_region)
        else:
            if not self.aws_access_key_id or not self.aws_secret_access_key:
                log_error(
                    "AWS credentials not found. Please set AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables or provide a boto3 session."
                )

            self.client = AwsClient(
                service_name="bedrock-runtime",
                region_name=self.aws_region,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
            )
        return self.client

    def get_async_client(self):
        """
        Get the async Bedrock client context manager.

        Returns:
            The async Bedrock client context manager.
        """
        if not AIOBOTO3_AVAILABLE:
            raise ImportError(
                "`aioboto3` not installed. Please install using `pip install aioboto3` for async support."
            )

        # When using a boto3 session, create the aioboto3 session from it
        # so that session credentials (IAM roles, EKS, STS) are respected.
        if self.session:
            credentials = self.session.get_credentials()
            if credentials is None:
                raise ValueError(
                    "boto3 session has no credentials. Check your AWS configuration "
                    "(environment variables, config files, IAM role, etc.)."
                )
            # Use local variables (not self.async_session) so concurrent
            # callers each get their own session and client.
            frozen = credentials.get_frozen_credentials()
            async_session = aioboto3.Session(
                aws_access_key_id=frozen.access_key,
                aws_secret_access_key=frozen.secret_key,
                aws_session_token=frozen.token,
                region_name=self.aws_region or self.session.region_name,
            )
            return async_session.client("bedrock-runtime")

        if self.async_session is None:
            self.aws_access_key_id = self.aws_access_key_id or getenv("AWS_ACCESS_KEY_ID")
            self.aws_secret_access_key = self.aws_secret_access_key or getenv("AWS_SECRET_ACCESS_KEY")
            self.aws_session_token = self.aws_session_token or getenv("AWS_SESSION_TOKEN")
            self.aws_region = self.aws_region or getenv("AWS_REGION")

            self.async_session = aioboto3.Session()

        client_kwargs: Dict[str, Any] = {
            "service_name": "bedrock-runtime",
            "region_name": self.aws_region,
        }

        if self.aws_sso_auth:
            pass
        else:
            if not self.aws_access_key_id or not self.aws_secret_access_key:
                import os

                env_access_key = os.environ.get("AWS_ACCESS_KEY_ID")
                env_secret_key = os.environ.get("AWS_SECRET_ACCESS_KEY")
                env_session_token = os.environ.get("AWS_SESSION_TOKEN")
                env_region = os.environ.get("AWS_REGION")

                if env_access_key and env_secret_key:
                    self.aws_access_key_id = env_access_key
                    self.aws_secret_access_key = env_secret_key
                    self.aws_session_token = env_session_token
                    if env_region:
                        self.aws_region = env_region
                        client_kwargs["region_name"] = self.aws_region

            if self.aws_access_key_id and self.aws_secret_access_key:
                client_kwargs["aws_access_key_id"] = self.aws_access_key_id
                client_kwargs["aws_secret_access_key"] = self.aws_secret_access_key
                if self.aws_session_token:
                    client_kwargs["aws_session_token"] = self.aws_session_token

        return self.async_session.client(**client_kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the model to a dictionary.

        Returns:
            Dict[str, Any]: The dictionary representation of the model.
        """
        model_dict = super().to_dict()
        model_dict.update(
            {
                "aws_region": self.aws_region,
                "max_tokens": self.max_tokens,
                "temperature": self.temperature,
                "top_p": self.top_p,
                "stop_sequences": self.stop_sequences,
            }
        )
        cleaned_dict = {k: v for k, v in model_dict.items() if v is not None}
        return cleaned_dict

    def _format_tools_for_request(self, tools: Optional[List[Dict[str, Any]]]) -> List[Dict[str, Any]]:
        """
        Format the tools for the request.

        Returns:
            List[Dict[str, Any]]: The formatted tools.
        """
        parsed_tools = []
        if tools is not None:
            for tool_def in tools:
                func_def = tool_def.get("function", {})
                properties = {}
                required = []

                for param_name, param_info in func_def.get("parameters", {}).get("properties", {}).items():
                    properties[param_name] = param_info.copy()

                    if "description" not in properties[param_name]:
                        properties[param_name]["description"] = ""

                    if "null" not in (
                        param_info.get("type") if isinstance(param_info.get("type"), list) else [param_info.get("type")]
                    ):
                        required.append(param_name)

                parsed_tools.append(
                    {
                        "toolSpec": {
                            "name": func_def.get("name") or "",
                            "description": func_def.get("description") or "",
                            "inputSchema": {"json": {"type": "object", "properties": properties, "required": required}},
                        }
                    }
                )

        return parsed_tools

    def _get_inference_config(self) -> Dict[str, Any]:
        """
        Get the inference config.

        Returns:
            Dict[str, Any]: The inference config.
        """
        request_kwargs = {
            "maxTokens": self.max_tokens,
            "temperature": self.temperature,
            "topP": self.top_p,
            "stopSequences": self.stop_sequences,
        }

        return {k: v for k, v in request_kwargs.items() if v is not None}

    def _format_messages(
        self, messages: List[Message], compress_tool_results: bool = False
    ) -> Tuple[List[Dict[str, Any]], Optional[List[Dict[str, Any]]]]:
        """
        Format the messages for the request.

        Args:
            messages: List of messages to format
            compress_tool_results: Whether to compress tool results

        Returns:
            Tuple[List[Dict[str, Any]], Optional[List[Dict[str, Any]]]]: The formatted messages.
        """
        from agno.utils.message import normalize_tool_messages

        # Backwards compat: expand old Gemini combined tool messages into individual canonical messages
        messages = normalize_tool_messages(messages)

        formatted_messages: List[Dict[str, Any]] = []
        system_message = None
        for message in messages:
            if message.role == "system":
                system_message = [{"text": message.content}]
            elif message.role == "tool":
                content = message.get_content(use_compressed_content=compress_tool_results)
                tool_result = {
                    "toolUseId": message.tool_call_id,
                    "content": [{"json": {"result": content}}],
                }
                if (
                    formatted_messages
                    and formatted_messages[-1]["role"] == "user"
                    and formatted_messages[-1]["content"]
                    and "toolResult" in formatted_messages[-1]["content"][0]
                ):
                    formatted_messages[-1]["content"].append({"toolResult": tool_result})
                else:
                    formatted_message: Dict[str, Any] = {"role": "user", "content": [{"toolResult": tool_result}]}
                    formatted_messages.append(formatted_message)
            else:
                formatted_message = {"role": message.role, "content": []}
                if isinstance(message.content, list):
                    formatted_message["content"].extend(message.content)
                elif message.tool_calls:
                    tool_use_content = []
                    for tool_call in message.tool_calls:
                        try:
                            # Parse arguments with error handling for empty or invalid JSON
                            arguments = tool_call["function"]["arguments"]
                            if not arguments or arguments.strip() == "":
                                tool_input = {}
                            else:
                                tool_input = json.loads(arguments)
                        except (json.JSONDecodeError, KeyError) as e:
                            log_warning(f"Failed to parse tool call arguments: {e}")
                            tool_input = {}

                        tool_use_content.append(
                            {
                                "toolUse": {
                                    "toolUseId": tool_call["id"],
                                    "name": tool_call["function"]["name"],
                                    "input": tool_input,
                                }
                            }
                        )
                    formatted_message["content"].extend(tool_use_content)
                else:
                    formatted_message["content"].append({"text": message.content})

                if message.images:
                    for image in message.images:
                        if not image.content:
                            raise ValueError("Image content is required for AWS Bedrock.")
                        if not image.format:
                            raise ValueError("Image format is required for AWS Bedrock.")

                        if image.format not in BEDROCK_SUPPORTED_IMAGE_FORMATS:
                            raise ValueError(
                                f"Unsupported image format: {image.format}. "
                                f"Supported formats: {BEDROCK_SUPPORTED_IMAGE_FORMATS}"
                            )

                        formatted_message["content"].append(
                            {
                                "image": {
                                    "format": image.format,
                                    "source": {
                                        "bytes": image.content,
                                    },
                                }
                            }
                        )
                if message.audio:
                    log_warning("Audio input is currently unsupported.")

                if message.videos:
                    for video in message.videos:
                        if not video.content:
                            raise ValueError("Video content is required for AWS Bedrock.")
                        if not video.format:
                            raise ValueError("Video format is required for AWS Bedrock.")

                        if video.format not in BEDROCK_SUPPORTED_VIDEO_FORMATS:
                            raise ValueError(
                                f"Unsupported video format: {video.format}. "
                                f"Supported formats: {BEDROCK_SUPPORTED_VIDEO_FORMATS}"
                            )

                        formatted_message["content"].append(
                            {
                                "video": {
                                    "format": video.format,
                                    "source": {
                                        "bytes": video.content,
                                    },
                                }
                            }
                        )

                if message.files:
                    for file in message.files:
                        if not file.content:
                            raise ValueError("File content is required for AWS Bedrock document input.")
                        if not file.format:
                            raise ValueError("File format is required for AWS Bedrock document input.")
                        if not file.name:
                            raise ValueError("File name is required for AWS Bedrock document input.")

                        if file.format not in BEDROCK_SUPPORTED_FILE_FORMATS:
                            raise ValueError(
                                f"Unsupported file format: {file.format}. "
                                f"Supported formats: {BEDROCK_SUPPORTED_FILE_FORMATS}"
                            )

                        formatted_message["content"].append(
                            {
                                "document": {
                                    "format": file.format,
                                    "name": file.name,
                                    "source": {
                                        "bytes": file.content,
                                    },
                                }
                            }
                        )

                formatted_messages.append(formatted_message)
        # TODO: Add caching: https://docs.aws.amazon.com/bedrock/latest/userguide/conversation-inference-call.html
        return formatted_messages, system_message

    def count_tokens(
        self,
        messages: List[Message],
        tools: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Union[Dict, Type[BaseModel]]] = None,
    ) -> int:
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results=True)
            converse_input: Dict[str, Any] = {"messages": formatted_messages}
            if system_message:
                converse_input["system"] = system_message

            response = self.get_client().count_tokens(modelId=self.id, input={"converse": converse_input})
            tokens = response.get("inputTokens", 0)

            # Count tool tokens
            if tools:
                from agno.utils.tokens import count_tool_tokens

                tokens += count_tool_tokens(tools, self.id)

            # Count schema tokens
            tokens += count_schema_tokens(output_schema, self.id)

            return tokens
        except Exception as e:
            log_warning(f"Failed to count tokens via Bedrock API: {e}")
            return super().count_tokens(messages, tools, output_schema)

    async def acount_tokens(
        self,
        messages: List[Message],
        tools: Optional[List[Dict[str, Any]]] = None,
        output_schema: Optional[Union[Dict, Type[BaseModel]]] = None,
    ) -> int:
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results=True)
            converse_input: Dict[str, Any] = {"messages": formatted_messages}
            if system_message:
                converse_input["system"] = system_message

            async with self.get_async_client() as client:
                response = await client.count_tokens(modelId=self.id, input={"converse": converse_input})
            tokens = response.get("inputTokens", 0)

            # Count tool tokens
            if tools:
                from agno.utils.tokens import count_tool_tokens

                tokens += count_tool_tokens(tools, self.id)

            # Count schema tokens
            tokens += count_schema_tokens(output_schema, self.id)

            return tokens
        except Exception as e:
            log_warning(f"Failed to count tokens via Bedrock API: {e}")
            return await super().acount_tokens(messages, tools, output_schema)

    def invoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> ModelResponse:
        """
        Invoke the Bedrock API.
        """
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results)

            tool_config = None
            if tools:
                tool_config = {"tools": self._format_tools_for_request(tools)}

            body = {
                "system": system_message,
                "toolConfig": tool_config,
                "inferenceConfig": self._get_inference_config(),
            }
            body = {k: v for k, v in body.items() if v is not None}

            if self.request_params:
                log_debug(f"Calling {self.provider} with request parameters: {self.request_params}", log_level=2)
                body.update(**self.request_params)

            assistant_message.metrics.start_timer()
            response = self.get_client().converse(modelId=self.id, messages=formatted_messages, **body)
            assistant_message.metrics.stop_timer()

            model_response = self._parse_provider_response(response, response_format=response_format)

            return model_response

        except ClientError as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e.response), model_name=self.name, model_id=self.id) from e
        except Exception as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    def invoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> Iterator[ModelResponse]:
        """
        Invoke the Bedrock API with streaming.
        """
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results)

            tool_config = None
            if tools:
                tool_config = {"tools": self._format_tools_for_request(tools)}

            body = {
                "system": system_message,
                "toolConfig": tool_config,
                "inferenceConfig": self._get_inference_config(),
            }
            body = {k: v for k, v in body.items() if v is not None}

            if self.request_params:
                body.update(**self.request_params)

            assistant_message.metrics.start_timer()

            # Track current tool being built across chunks
            current_tool: Dict[str, Any] = {}

            for chunk in self.get_client().converse_stream(modelId=self.id, messages=formatted_messages, **body)[
                "stream"
            ]:
                model_response, current_tool = self._parse_provider_response_delta(chunk, current_tool)
                yield model_response

            assistant_message.metrics.stop_timer()

        except ClientError as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e.response), model_name=self.name, model_id=self.id) from e
        except Exception as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> ModelResponse:
        """
        Async invoke the Bedrock API.
        """
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results)

            tool_config = None
            if tools:
                tool_config = {"tools": self._format_tools_for_request(tools)}

            body = {
                "system": system_message,
                "toolConfig": tool_config,
                "inferenceConfig": self._get_inference_config(),
            }
            body = {k: v for k, v in body.items() if v is not None}

            if self.request_params:
                log_debug(f"Calling {self.provider} with request parameters: {self.request_params}", log_level=2)
                body.update(**self.request_params)

            assistant_message.metrics.start_timer()

            async with self.get_async_client() as client:
                response = await client.converse(modelId=self.id, messages=formatted_messages, **body)

            assistant_message.metrics.stop_timer()

            model_response = self._parse_provider_response(response, response_format=response_format)

            return model_response

        except ClientError as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e.response), model_name=self.name, model_id=self.id) from e
        except Exception as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
        compress_tool_results: bool = False,
    ) -> AsyncIterator[ModelResponse]:
        """
        Async invoke the Bedrock API with streaming.
        """
        try:
            formatted_messages, system_message = self._format_messages(messages, compress_tool_results)

            tool_config = None
            if tools:
                tool_config = {"tools": self._format_tools_for_request(tools)}

            body = {
                "system": system_message,
                "toolConfig": tool_config,
                "inferenceConfig": self._get_inference_config(),
            }
            body = {k: v for k, v in body.items() if v is not None}

            if self.request_params:
                body.update(**self.request_params)

            assistant_message.metrics.start_timer()

            # Track current tool being built across chunks
            current_tool: Dict[str, Any] = {}

            async with self.get_async_client() as client:
                response = await client.converse_stream(modelId=self.id, messages=formatted_messages, **body)
                async for chunk in response["stream"]:
                    model_response, current_tool = self._parse_provider_response_delta(chunk, current_tool)
                    yield model_response

            assistant_message.metrics.stop_timer()

        except ClientError as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e.response), model_name=self.name, model_id=self.id) from e
        except Exception as e:
            log_error(f"Unexpected error calling Bedrock API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    # Overwrite the default from the base model
    def format_function_call_results(
        self,
        messages: List[Message],
        function_call_results: List[Message],
        compress_tool_results: bool = False,
        **kwargs,
    ) -> None:
        """
        Handle the results of function calls for Bedrock.
        Uses compressed_content if compress_tool_results is True.

        Args:
            messages (List[Message]): The list of conversation messages.
            function_call_results (List[Message]): The results of the function calls.
            compress_tool_results: Whether to compress tool results.
            **kwargs: Additional arguments including tool_ids.
        """

        if function_call_results:
            tool_ids = kwargs.get("tool_ids", [])

            for _fc_message_index, _fc_message in enumerate(function_call_results):
                # Use tool_call_id from message if tool_ids list is insufficient
                tool_id = tool_ids[_fc_message_index] if _fc_message_index < len(tool_ids) else _fc_message.tool_call_id
                if not _fc_message.tool_call_id:
                    _fc_message.tool_call_id = tool_id

                # Append as standard role="tool" message
                messages.append(_fc_message)

    def _parse_provider_response(self, response: Dict[str, Any], **kwargs) -> ModelResponse:
        """
        Parse the provider response.

        Args:
            response (Dict[str, Any]): The response from the provider.

        Returns:
            ModelResponse: The parsed response.
        """
        model_response = ModelResponse()

        if "output" in response and "message" in response["output"]:
            message = response["output"]["message"]
            # Set the role of the message
            model_response.role = message["role"]

            # Get the content of the message
            content = message["content"]

            # Tools
            if "stopReason" in response and response["stopReason"] == "tool_use":
                model_response.tool_calls = []
                model_response.extra = model_response.extra or {}
                model_response.extra["tool_ids"] = []
                for tool in content:
                    if "toolUse" in tool:
                        model_response.extra["tool_ids"].append(tool["toolUse"]["toolUseId"])
                        model_response.tool_calls.append(
                            {
                                "id": tool["toolUse"]["toolUseId"],
                                "type": "function",
                                "function": {
                                    "name": tool["toolUse"]["name"],
                                    "arguments": json.dumps(tool["toolUse"]["input"]),
                                },
                            }
                        )

            # Extract text content if it's a list of dictionaries
            if isinstance(content, list) and content and isinstance(content[0], dict):
                content = [item.get("text", "") for item in content if "text" in item]
                content = "\n".join(content)  # Join multiple text items if present

            model_response.content = content

        if "usage" in response:
            model_response.response_usage = self._get_metrics(response["usage"])

        return model_response

    def _parse_provider_response_delta(
        self, response_delta: Dict[str, Any], current_tool: Dict[str, Any]
    ) -> Tuple[ModelResponse, Dict[str, Any]]:
        """Parse the provider response delta for streaming.

        Args:
            response_delta: The streaming response delta from AWS Bedrock
            current_tool: The current tool being built across chunks

        Returns:
            Tuple[ModelResponse, Dict[str, Any]]: The parsed model response delta and updated current_tool
        """
        model_response = ModelResponse(role="assistant")

        # Handle contentBlockStart - tool use start
        if "contentBlockStart" in response_delta:
            start = response_delta["contentBlockStart"]["start"]
            if "toolUse" in start:
                # Start a new tool
                tool_use_data = start["toolUse"]
                current_tool = {
                    "id": tool_use_data.get("toolUseId", ""),
                    "type": "function",
                    "function": {
                        "name": tool_use_data.get("name", ""),
                        "arguments": "",  # Will be filled in subsequent deltas
                    },
                }

        # Handle contentBlockDelta - text content or tool input
        elif "contentBlockDelta" in response_delta:
            delta = response_delta["contentBlockDelta"]["delta"]
            if "text" in delta:
                model_response.content = delta["text"]
            elif "toolUse" in delta and current_tool:
                # Accumulate tool input
                tool_input = delta["toolUse"].get("input", "")
                if tool_input:
                    current_tool["function"]["arguments"] += tool_input

        # Handle contentBlockStop - tool use complete
        elif "contentBlockStop" in response_delta and current_tool:
            # Tool is complete, add it to model response
            model_response.tool_calls = [current_tool]
            # Track tool_id in extra for format_function_call_results
            model_response.extra = {"tool_ids": [current_tool["id"]]}
            # Reset current_tool for next tool
            current_tool = {}

        # Handle metadata/usage information
        elif "metadata" in response_delta or "messageStop" in response_delta:
            body = response_delta.get("metadata") or response_delta.get("messageStop") or {}
            if "usage" in body:
                model_response.response_usage = self._get_metrics(body["usage"])

        return model_response, current_tool

    def _get_metrics(self, response_usage: Dict[str, Any]) -> MessageMetrics:
        """
        Parse the given AWS Bedrock usage into an Agno MessageMetrics object.

        Args:
            response_usage: Usage data from AWS Bedrock

        Returns:
            MessageMetrics: Parsed metrics data
        """
        metrics = MessageMetrics()

        metrics.input_tokens = response_usage.get("inputTokens", 0) or 0
        metrics.output_tokens = response_usage.get("outputTokens", 0) or 0
        metrics.total_tokens = metrics.input_tokens + metrics.output_tokens

        metrics.cache_read_tokens = response_usage.get("cacheReadInputTokens", 0) or 0
        metrics.cache_write_tokens = response_usage.get("cacheWriteInputTokens", 0) or 0

        return metrics



import base64
import time
from os import getenv
from typing import Any, Optional
from uuid import uuid4

from agno.agent import Agent
from agno.media import Image, Video
from agno.tools import Toolkit
from agno.tools.function import ToolResult
from agno.utils.log import log_debug, log_error, log_info

try:
    from google.genai import Client
    from google.genai.types import GenerateImagesResponse, GenerateVideosOperation
except (ModuleNotFoundError, ImportError):
    raise ImportError("`google-genai` not installed. Please install using `pip install google-genai`")


class GeminiTools(Toolkit):
    """Tools for interacting with Google Gemini API"""

    def __init__(
        self,
        api_key: Optional[str] = None,
        vertexai: bool = False,
        project_id: Optional[str] = None,
        location: Optional[str] = None,
        image_generation_model: str = "imagen-3.0-generate-002",
        video_generation_model: str = "veo-2.0-generate-001",
        enable_generate_image: bool = True,
        enable_generate_video: bool = True,
        all: bool = False,
        **kwargs,
    ):
        tools = []
        if all or enable_generate_image:
            tools.append(self.generate_image)
        if all or enable_generate_video:
            tools.append(self.generate_video)

        super().__init__(name="gemini_tools", tools=tools, **kwargs)

        # Set mode and credentials: use only provided vertexai parameter
        self.vertexai = vertexai or getenv("GOOGLE_GENAI_USE_VERTEXAI") == "true"
        self.project_id = project_id
        self.location = location

        # Load API key from argument or environment
        self.api_key = api_key or getenv("GOOGLE_API_KEY")
        if not self.vertexai and not self.api_key:
            log_error("GOOGLE_API_KEY not set. Please set the GOOGLE_API_KEY environment variable.")
            raise ValueError("GOOGLE_API_KEY not set. Please provide api_key or set the environment variable.")

        # Prepare client parameters
        client_params: dict[str, Any] = {}
        if self.vertexai:
            log_info("Using Vertex AI API")
            client_params["vertexai"] = True
            client_params["project"] = self.project_id or getenv("GOOGLE_CLOUD_PROJECT")
            client_params["location"] = self.location or getenv("GOOGLE_CLOUD_LOCATION")
        else:
            log_info("Using Gemini API")
            client_params["api_key"] = self.api_key

        try:
            self.client = Client(**client_params)
            log_debug("Google GenAI Client created successfully.")
        except Exception as e:
            log_error(f"Failed to create Google GenAI Client: {e}", exc_info=True)
            raise ValueError(f"Failed to create Google GenAI Client. Error: {e}")

        self.image_model = image_generation_model
        self.video_model = video_generation_model

    def generate_image(
        self,
        agent: Agent,
        prompt: str,
    ) -> ToolResult:
        """Generate images based on a text prompt using Google Imagen.

        Args:
            prompt (str): The text prompt to generate the image from.
        Returns:
            ToolResult: A ToolResult containing the generated images or error message.
        """

        try:
            response: GenerateImagesResponse = self.client.models.generate_images(
                model=self.image_model,
                prompt=prompt,
            )

            log_debug("DEBUG: Raw Gemini API response")

            # Extract image bytes
            if response.generated_images is None or not response.generated_images:
                log_info("No images were generated.")
                return ToolResult(content="Failed to generate image: No images were generated.")

            generated_images = []
            for generated_image in response.generated_images:
                if generated_image.image is None or not generated_image.image.image_bytes:
                    continue

                image_bytes = generated_image.image.image_bytes
                actual_mime_type = "image/png"
                media_id = str(uuid4())

                # Create ImageArtifact with raw bytes (not base64 encoded)
                image_artifact = Image(
                    id=media_id,
                    content=image_bytes,
                    original_prompt=prompt,
                    mime_type=actual_mime_type,
                )
                generated_images.append(image_artifact)
                log_debug(f"Successfully generated image {media_id} with model {self.image_model}")

            if generated_images:
                return ToolResult(
                    content="Image generated successfully",
                    images=generated_images,
                )
            else:
                return ToolResult(content="Failed to generate image: No valid images were generated.")

        except Exception as e:
            log_error(f"Failed to generate image: Client or method not available ({e})")
            return ToolResult(content=f"Failed to generate image: Client or method not available ({e})")

    def generate_video(
        self,
        agent: Agent,
        prompt: str,
    ) -> ToolResult:
        """Generate a video based on a text prompt.
        Args:
            prompt (str): The text prompt to generate the video from.
        Returns:
            ToolResult: A ToolResult containing the generated video or error message.
        """
        # Video generation requires Vertex AI mode.
        if not self.vertexai:
            log_error("Video generation requires Vertex AI mode. Please enable Vertex AI mode.")
            return ToolResult(
                content="Video generation requires Vertex AI mode. "
                "Please set `vertexai=True` or environment variable `GOOGLE_GENAI_USE_VERTEXAI=true`."
            )

        from google.genai.types import GenerateVideosConfig

        try:
            operation: GenerateVideosOperation = self.client.models.generate_videos(
                model=self.video_model,
                prompt=prompt,
                config=GenerateVideosConfig(
                    enhance_prompt=True,
                ),
            )

            while not operation.done:
                time.sleep(5)
                operation = self.client.operations.get(operation=operation)

            result = operation.result
            if result is None or result.generated_videos is None or not result.generated_videos:
                log_error("No videos were generated.")
                return ToolResult(content="Failed to generate video: No videos were generated.")

            generated_videos = []
            for video in result.generated_videos:
                if video.video is None or not video.video.video_bytes:
                    continue

                generated_video = video.video
                if generated_video.video_bytes is None:
                    continue

                media_id = str(uuid4())

                # Create VideoArtifact with base64 encoded content
                video_artifact = Video(
                    id=media_id,
                    content=base64.b64encode(generated_video.video_bytes).decode("utf-8"),
                    original_prompt=prompt,
                    mime_type=generated_video.mime_type or "video/mp4",
                )
                generated_videos.append(video_artifact)
                log_debug(f"Successfully generated video {media_id} with model {self.video_model}")

            if generated_videos:
                return ToolResult(
                    content="Video generated successfully",
                    videos=generated_videos,
                )
            else:
                return ToolResult(content="Failed to generate video: No valid videos were generated.")

        except Exception as e:
            log_error(f"Failed to generate video: {e}")
            return ToolResult(content=f"Failed to generate video: {e}")



from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

from pydantic import BaseModel

from agno.models.message import Message
from agno.run.agent import RunOutput, RunStatus
from agno.run.team import TeamRunOutput
from agno.session.summary import SessionSummary
from agno.utils.log import log_debug, log_warning


@dataclass
class TeamSession:
    """Team Session that is stored in the database"""

    # Session UUID
    session_id: str

    # ID of the team that this session is associated with
    team_id: Optional[str] = None
    # ID of the user interacting with this team
    user_id: Optional[str] = None
    # ID of the workflow that this session is associated with
    workflow_id: Optional[str] = None

    # Team Data: agent_id, name and model
    team_data: Optional[Dict[str, Any]] = None
    # Session Data: session_name, session_state, images, videos, audio
    session_data: Optional[Dict[str, Any]] = None
    # Metadata stored with this team
    metadata: Optional[Dict[str, Any]] = None
    # List of all runs in the session
    runs: Optional[list[Union[TeamRunOutput, RunOutput]]] = None
    # Summary of the session
    summary: Optional[SessionSummary] = None

    # The unix timestamp when this session was created
    created_at: Optional[int] = None
    # The unix timestamp when this session was last updated
    updated_at: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        session_dict = asdict(self)

        session_dict["runs"] = [run.to_dict() for run in self.runs] if self.runs else None
        session_dict["summary"] = self.summary.to_dict() if self.summary else None

        return session_dict

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Optional[TeamSession]:
        if data is None or data.get("session_id") is None:
            log_warning("TeamSession is missing session_id")
            return None

        summary = data.get("summary")
        if summary is not None and isinstance(summary, dict):
            data["summary"] = SessionSummary.from_dict(data["summary"])  # type: ignore

        runs = data.get("runs")
        serialized_runs: List[Union[TeamRunOutput, RunOutput]] = []
        if runs is not None and isinstance(runs[0], dict):
            for run in runs:
                if "agent_id" in run:
                    serialized_runs.append(RunOutput.from_dict(run))
                elif "team_id" in run:
                    serialized_runs.append(TeamRunOutput.from_dict(run))

        return cls(
            session_id=data.get("session_id"),  # type: ignore
            team_id=data.get("team_id"),
            user_id=data.get("user_id"),
            workflow_id=data.get("workflow_id"),
            team_data=data.get("team_data"),
            session_data=data.get("session_data"),
            metadata=data.get("metadata"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            runs=serialized_runs,
            summary=data.get("summary"),
        )

    def get_run(self, run_id: str) -> Optional[Union[TeamRunOutput, RunOutput]]:
        for run in self.runs or []:
            if run.run_id == run_id:
                return run
        return None

    def upsert_run(self, run_response: Union[TeamRunOutput, RunOutput]):
        """Adds a RunOutput, together with some calculated data, to the runs list."""
        messages = run_response.messages

        # Clear message timer before storage
        for m in messages or []:
            if m.metrics is not None and hasattr(m.metrics, "timer"):
                m.metrics.timer = None

        if not self.runs:
            self.runs = []

        for i, existing_run in enumerate(self.runs or []):
            if existing_run.run_id == run_response.run_id:
                self.runs[i] = run_response
                break
        else:
            self.runs.append(run_response)

        log_debug("Added RunOutput to Team Session")

    def get_messages(
        self,
        team_id: Optional[str] = None,
        member_ids: Optional[List[str]] = None,
        last_n_runs: Optional[int] = None,
        limit: Optional[int] = None,
        skip_roles: Optional[List[str]] = None,
        skip_statuses: Optional[List[RunStatus]] = None,
        skip_history_messages: bool = True,
        skip_member_messages: bool = True,
    ) -> List[Message]:
        """Returns the messages belonging to the session that fit the given criteria.

        Args:
            team_id: The id of the team to get the messages from.
            member_ids: The ids of the members to get the messages from.
            last_n_runs: The number of runs to return messages from, counting from the latest. Defaults to all runs.
            limit: The number of messages to return, counting from the latest. Defaults to all messages.
            skip_roles: Skip messages with these roles.
            skip_statuses: Skip messages with these statuses.
            skip_history_messages: Skip messages that were tagged as history in previous runs.
            skip_member_messages: Skip messages created by members of the team.

        Returns:
            A list of Messages belonging to the session.
        """

        def _should_skip_message(
            message: Message, skip_roles: Optional[List[str]] = None, skip_history_messages: bool = True
        ) -> bool:
            """Processes a message for history"""
            # Skip messages that were tagged as history in previous runs
            if hasattr(message, "from_history") and message.from_history and skip_history_messages:
                return True

            # Skip messages with specified role
            if skip_roles and message.role in skip_roles:
                return True
            return False

        if member_ids is not None and skip_member_messages:
            log_debug("Member IDs to filter by were provided. The skip_member_messages flag will be ignored.")
            skip_member_messages = False

        if not self.runs:
            return []

        if skip_statuses is None:
            skip_statuses = [RunStatus.paused, RunStatus.cancelled, RunStatus.error]

        session_runs = self.runs

        # Filter by team_id and member_ids
        if team_id:
            session_runs = [run for run in session_runs if hasattr(run, "team_id") and run.team_id == team_id]  # type: ignore
        if member_ids:
            filtered_runs = []
            for run in session_runs:
                if hasattr(run, "agent_id") and run.agent_id in member_ids:  # type: ignore
                    filtered_runs.append(run)
                elif hasattr(run, "member_responses"):
                    for member_run in run.member_responses:
                        if hasattr(member_run, "agent_id") and member_run.agent_id in member_ids:  # type: ignore
                            filtered_runs.append(member_run)
            session_runs = filtered_runs

        if skip_member_messages:
            # Filter for the top-level runs (main team runs or agent runs when sharing session)
            session_runs = [run for run in session_runs if run.parent_run_id is None]  # type: ignore

        # Filter by status
        session_runs = [run for run in session_runs if hasattr(run, "status") and run.status not in skip_statuses]  # type: ignore

        messages_from_history = []
        system_message = None

        # Limit the number of messages returned if limit is set
        if limit is not None:
            for run_response in session_runs:
                if not run_response or not run_response.messages:
                    continue

                for message in run_response.messages or []:
                    if _should_skip_message(message, skip_roles, skip_history_messages):
                        continue

                    if message.role == "system":
                        # Only add the system message once
                        if system_message is None:
                            system_message = message
                    else:
                        messages_from_history.append(message)

            if system_message:
                messages_from_history = [system_message] + messages_from_history[
                    -(limit - 1) :
                ]  # Grab one less message then add the system message
            else:
                messages_from_history = messages_from_history[-limit:]

            # Remove tool result messages that don't have an associated assistant message with tool calls
            while len(messages_from_history) > 0 and messages_from_history[0].role == "tool":
                messages_from_history.pop(0)
        else:
            # Filter by last_n runs
            runs_to_process = session_runs[-last_n_runs:] if last_n_runs is not None else session_runs

            for run_response in runs_to_process:
                if not (run_response and run_response.messages):
                    continue

                for message in run_response.messages or []:
                    if _should_skip_message(message, skip_roles, skip_history_messages):
                        continue

                    if message.role == "system":
                        # Only add the system message once
                        if system_message is None:
                            system_message = message
                            messages_from_history.append(system_message)
                    else:
                        messages_from_history.append(message)

        log_debug(f"Getting messages from previous runs: {len(messages_from_history)}")
        return messages_from_history

    def get_chat_history(self, last_n_runs: Optional[int] = None) -> List[Message]:
        """Return the chat history (user and assistant messages) for the session.
        Use get_messages() for more filtering options.

        Args:
            last_n_runs: Number of recent runs to include. If None, all runs will be considered.

        Returns:
            A list of user and assistant Messages belonging to the session.
        """
        return self.get_messages(skip_roles=["system", "tool"], skip_member_messages=True, last_n_runs=last_n_runs)

    def get_tool_calls(self, num_calls: Optional[int] = None) -> List[Dict[str, Any]]:
        """Returns a list of tool calls from the messages"""

        tool_calls = []
        session_runs = self.runs
        if session_runs is None:
            return []

        for run_response in session_runs[::-1]:
            if run_response and run_response.messages:
                for message in run_response.messages or []:
                    if message.tool_calls:
                        for tool_call in message.tool_calls:
                            tool_calls.append(tool_call)
                            if num_calls and len(tool_calls) >= num_calls:
                                return tool_calls
        return tool_calls

    def get_team_history(self, num_runs: Optional[int] = None) -> List[Tuple[str, str]]:
        """Get team history as structured data (input, response pairs) -> This is the history of the team leader, not the members.

        Args:
            num_runs: Number of recent runs to include. If None, returns all available history.
        """
        if not self.runs:
            return []

        from agno.run.base import RunStatus

        # Get completed runs only (exclude current/pending run)
        completed_runs = [run for run in self.runs if run.status == RunStatus.completed and run.parent_run_id is None]

        if num_runs is not None and len(completed_runs) > num_runs:
            recent_runs = completed_runs[-num_runs:]
        else:
            recent_runs = completed_runs

        if not recent_runs:
            return []

        # Return structured data as list of (input, response) tuples
        history_data = []
        for run in recent_runs:
            # Get input
            input_str = ""
            if run.input:
                input_str = run.input.input_content_string()

            # Get response
            response_str = ""
            if run.content:
                response_str = (
                    run.content.model_dump_json(indent=2, exclude_none=True)
                    if isinstance(run.content, BaseModel)
                    else str(run.content)
                )

            history_data.append((input_str, response_str))

        return history_data

    def get_team_history_context(self, num_runs: Optional[int] = None) -> Optional[str]:
        """Get formatted team history context for steps

        Args:
            num_runs: Number of recent runs to include. If None, returns all available history.
        """
        history_data = self.get_team_history(num_runs)

        if not history_data:
            return None

        # Format as team history context using the structured data
        context_parts = ["<team_history_context>"]

        for i, (input_str, response_str) in enumerate(history_data, 1):
            context_parts.append(f"[run-{i}]")

            if input_str:
                context_parts.append(f"input: {input_str}")
            if response_str:
                context_parts.append(f"response: {response_str}")

            context_parts.append("")  # Empty line between runs

        context_parts.append("</team_history_context>")
        context_parts.append("")  # Empty line before current input

        return "\n".join(context_parts)

    def get_session_summary(self) -> Optional[SessionSummary]:
        """Get the session summary for the session"""

        if self.summary is None:
            return None

        return self.summary  # type: ignore



"""SchedulerTools -- give agents the ability to create and manage recurring schedules.

Wraps the ScheduleManager to expose schedule CRUD as agent-callable tools.
Requires a database backend that implements the scheduler DB methods and
the AgentOS server + SchedulePoller to actually execute the schedules.

Example:
    from agno.tools.scheduler import SchedulerTools

    agent = Agent(
        model=OpenAIChat(id="gpt-4o"),
        tools=[
            SchedulerTools(
                db=scheduler_db,
                default_endpoint="/agents/my-agent/runs",
            )
        ],
    )
"""

import json
from typing import Any, Callable, Dict, List, Optional

from agno.scheduler.manager import ScheduleManager
from agno.tools.toolkit import Toolkit
from agno.utils.log import log_debug, logger


class SchedulerTools(Toolkit):
    """Toolkit that lets an agent create and manage recurring schedules.

    The agent can ask a user "what should I do every day?" and then call
    ``create_schedule`` to set up a cron-based recurring execution via the
    existing AgentOS scheduler infrastructure.

    Args:
        db: A database adapter that implements the scheduler DB methods.
        default_endpoint: The default endpoint to call when a schedule fires
            (e.g. ``/agents/<agent_id>/runs``). The agent can override this
            per-schedule, but having a default simplifies the common case.
        default_method: HTTP method for the endpoint (default: ``POST``).
        default_timezone: Default timezone for schedules (default: ``UTC``).
        default_payload: Default payload to send with each scheduled run.
            The agent can override or extend this per-schedule.
    """

    def __init__(
        self,
        db: Any,
        default_endpoint: Optional[str] = None,
        default_method: str = "POST",
        default_timezone: str = "UTC",
        default_payload: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        self.manager = ScheduleManager(db=db)
        self.default_endpoint = default_endpoint
        self.default_method = default_method
        self.default_timezone = default_timezone
        self.default_payload = default_payload

        tools: List[Callable] = [
            self.create_schedule,
            self.list_schedules,
            self.get_schedule,
            self.delete_schedule,
            self.enable_schedule,
            self.disable_schedule,
            self.get_schedule_runs,
        ]

        async_tools: List[tuple[Callable[..., Any], str]] = [
            (self.acreate_schedule, "create_schedule"),
            (self.alist_schedules, "list_schedules"),
            (self.aget_schedule, "get_schedule"),
            (self.adelete_schedule, "delete_schedule"),
            (self.aenable_schedule, "enable_schedule"),
            (self.adisable_schedule, "disable_schedule"),
            (self.aget_schedule_runs, "get_schedule_runs"),
        ]

        super().__init__(
            name="scheduler",
            tools=tools,
            async_tools=async_tools,
            instructions=(
                "Use these tools to create and manage recurring scheduled tasks. "
                "When a user asks you to do something on a recurring basis (daily, weekly, etc.), "
                "convert their request into a cron expression and create a schedule. "
                "Common cron patterns: '0 9 * * *' (daily at 9am), '0 9 * * 1' (every Monday at 9am), "
                "'0 */6 * * *' (every 6 hours), '0 9 * * 1-5' (weekdays at 9am). "
                "When creating schedules for run endpoints (ending in /runs), you MUST include a "
                '\'message\' field in the payload, e.g. {"message": "Run the daily health check"}.'
            ),
            **kwargs,
        )

    @staticmethod
    def _is_run_endpoint(endpoint: str, method: str) -> bool:
        """Check if the endpoint targets an agent/team/workflow run route."""
        return method.upper() == "POST" and endpoint.rstrip("/").endswith("/runs")

    # ------------------------------------------------------------------
    # Sync tools
    # ------------------------------------------------------------------

    def create_schedule(
        self,
        name: str,
        cron: str,
        description: Optional[str] = None,
        endpoint: Optional[str] = None,
        method: Optional[str] = None,
        payload: Optional[str] = None,
        timezone: Optional[str] = None,
    ) -> str:
        """Create a new recurring schedule that runs on a cron expression.

        Args:
            name (str): A unique name for the schedule (e.g. "daily-weather-report").
            cron (str): A 5-field cron expression (e.g. "0 9 * * *" for daily at 9am UTC).
            description (str): A human-readable description of what this schedule does.
            endpoint (str): The API endpoint to call when the schedule fires. Uses the default if not provided.
            method (str): HTTP method (GET, POST, etc.). Uses the default if not provided.
            payload (str): JSON string of the payload to send with the request. Uses the default if not provided.
            timezone (str): Timezone for the cron expression (e.g. "America/New_York"). Uses the default if not provided.

        Returns:
            str: JSON string with the created schedule details.
        """
        resolved_endpoint = endpoint or self.default_endpoint
        if not resolved_endpoint:
            return json.dumps({"error": "No endpoint provided and no default_endpoint configured"})

        resolved_payload = self.default_payload
        if payload is not None:
            try:
                resolved_payload = json.loads(payload)
            except json.JSONDecodeError:
                return json.dumps({"error": "Invalid JSON in payload parameter"})

        # Run endpoints require a "message" field in the payload
        resolved_method = method or self.default_method
        if self._is_run_endpoint(resolved_endpoint, resolved_method):
            if not resolved_payload or "message" not in resolved_payload:
                return json.dumps(
                    {
                        "error": "Schedules targeting run endpoints require a 'message' field in the payload. "
                        'Provide payload with at least: {"message": "your prompt here"}'
                    }
                )

        try:
            schedule = self.manager.create(
                name=name,
                cron=cron,
                endpoint=resolved_endpoint,
                method=resolved_method,
                description=description,
                payload=resolved_payload,
                timezone=timezone or self.default_timezone,
                if_exists="update",
            )
            log_debug(f"Schedule created: {schedule.name} ({schedule.cron_expr})")
            return json.dumps(
                {
                    "status": "created",
                    "id": schedule.id,
                    "name": schedule.name,
                    "cron": schedule.cron_expr,
                    "endpoint": schedule.endpoint,
                    "timezone": schedule.timezone,
                    "enabled": schedule.enabled,
                    "description": schedule.description,
                }
            )
        except Exception as e:
            logger.error(f"Failed to create schedule: {e}")
            return json.dumps({"error": str(e)})

    def list_schedules(self, enabled_only: bool = False) -> str:
        """List all existing schedules.

        Args:
            enabled_only (bool): If True, only return enabled schedules. Defaults to False.

        Returns:
            str: JSON string with the list of schedules.
        """
        try:
            enabled_filter = True if enabled_only else None
            schedules = self.manager.list(enabled=enabled_filter)
            result = [
                {
                    "id": s.id,
                    "name": s.name,
                    "cron": s.cron_expr,
                    "endpoint": s.endpoint,
                    "timezone": s.timezone,
                    "enabled": s.enabled,
                    "description": s.description,
                }
                for s in schedules
            ]
            return json.dumps({"schedules": result, "count": len(result)})
        except Exception as e:
            logger.error(f"Failed to list schedules: {e}")
            return json.dumps({"error": str(e)})

    def get_schedule(self, schedule_id: str) -> str:
        """Get details of a specific schedule by its ID.

        Args:
            schedule_id (str): The ID of the schedule to retrieve.

        Returns:
            str: JSON string with the schedule details.
        """
        try:
            schedule = self.manager.get(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "id": schedule.id,
                    "name": schedule.name,
                    "cron": schedule.cron_expr,
                    "endpoint": schedule.endpoint,
                    "method": schedule.method,
                    "timezone": schedule.timezone,
                    "enabled": schedule.enabled,
                    "description": schedule.description,
                    "payload": schedule.payload,
                }
            )
        except Exception as e:
            logger.error(f"Failed to get schedule: {e}")
            return json.dumps({"error": str(e)})

    def delete_schedule(self, schedule_id: str) -> str:
        """Delete a schedule by its ID. This permanently removes the schedule.

        Args:
            schedule_id (str): The ID of the schedule to delete.

        Returns:
            str: JSON string confirming deletion.
        """
        try:
            deleted = self.manager.delete(schedule_id)
            if deleted:
                return json.dumps({"status": "deleted", "id": schedule_id})
            return json.dumps({"error": f"Schedule not found or could not be deleted: {schedule_id}"})
        except Exception as e:
            logger.error(f"Failed to delete schedule: {e}")
            return json.dumps({"error": str(e)})

    def enable_schedule(self, schedule_id: str) -> str:
        """Enable a disabled schedule so it starts running again.

        Args:
            schedule_id (str): The ID of the schedule to enable.

        Returns:
            str: JSON string with the updated schedule details.
        """
        try:
            schedule = self.manager.enable(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "status": "enabled",
                    "id": schedule.id,
                    "name": schedule.name,
                    "enabled": schedule.enabled,
                }
            )
        except Exception as e:
            logger.error(f"Failed to enable schedule: {e}")
            return json.dumps({"error": str(e)})

    def disable_schedule(self, schedule_id: str) -> str:
        """Disable a schedule so it stops running. Can be re-enabled later.

        Args:
            schedule_id (str): The ID of the schedule to disable.

        Returns:
            str: JSON string with the updated schedule details.
        """
        try:
            schedule = self.manager.disable(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "status": "disabled",
                    "id": schedule.id,
                    "name": schedule.name,
                    "enabled": schedule.enabled,
                }
            )
        except Exception as e:
            logger.error(f"Failed to disable schedule: {e}")
            return json.dumps({"error": str(e)})

    def get_schedule_runs(self, schedule_id: str, limit: int = 10) -> str:
        """Get the run history for a schedule.

        Args:
            schedule_id (str): The ID of the schedule to get runs for.
            limit (int): Maximum number of runs to return. Defaults to 10.

        Returns:
            str: JSON string with the list of runs.
        """
        try:
            runs = self.manager.get_runs(schedule_id, limit=limit)
            result = [
                {
                    "id": r.id,
                    "status": r.status,
                    "triggered_at": r.triggered_at,
                    "completed_at": r.completed_at,
                    "error": r.error,
                }
                for r in runs
            ]
            return json.dumps({"runs": result, "count": len(result)})
        except Exception as e:
            logger.error(f"Failed to get schedule runs: {e}")
            return json.dumps({"error": str(e)})

    # ------------------------------------------------------------------
    # Async tools
    # ------------------------------------------------------------------

    async def acreate_schedule(
        self,
        name: str,
        cron: str,
        description: Optional[str] = None,
        endpoint: Optional[str] = None,
        method: Optional[str] = None,
        payload: Optional[str] = None,
        timezone: Optional[str] = None,
    ) -> str:
        """Create a new recurring schedule that runs on a cron expression.

        Args:
            name (str): A unique name for the schedule (e.g. "daily-weather-report").
            cron (str): A 5-field cron expression (e.g. "0 9 * * *" for daily at 9am UTC).
            description (str): A human-readable description of what this schedule does.
            endpoint (str): The API endpoint to call when the schedule fires. Uses the default if not provided.
            method (str): HTTP method (GET, POST, etc.). Uses the default if not provided.
            payload (str): JSON string of the payload to send with the request. Uses the default if not provided.
            timezone (str): Timezone for the cron expression (e.g. "America/New_York"). Uses the default if not provided.

        Returns:
            str: JSON string with the created schedule details.
        """
        resolved_endpoint = endpoint or self.default_endpoint
        if not resolved_endpoint:
            return json.dumps({"error": "No endpoint provided and no default_endpoint configured"})

        resolved_payload = self.default_payload
        if payload is not None:
            try:
                resolved_payload = json.loads(payload)
            except json.JSONDecodeError:
                return json.dumps({"error": "Invalid JSON in payload parameter"})

        # Run endpoints require a "message" field in the payload
        resolved_method = method or self.default_method
        if self._is_run_endpoint(resolved_endpoint, resolved_method):
            if not resolved_payload or "message" not in resolved_payload:
                return json.dumps(
                    {
                        "error": "Schedules targeting run endpoints require a 'message' field in the payload. "
                        'Provide payload with at least: {"message": "your prompt here"}'
                    }
                )

        try:
            schedule = await self.manager.acreate(
                name=name,
                cron=cron,
                endpoint=resolved_endpoint,
                method=resolved_method,
                description=description,
                payload=resolved_payload,
                timezone=timezone or self.default_timezone,
                if_exists="update",
            )
            log_debug(f"Schedule created: {schedule.name} ({schedule.cron_expr})")
            return json.dumps(
                {
                    "status": "created",
                    "id": schedule.id,
                    "name": schedule.name,
                    "cron": schedule.cron_expr,
                    "endpoint": schedule.endpoint,
                    "timezone": schedule.timezone,
                    "enabled": schedule.enabled,
                    "description": schedule.description,
                }
            )
        except Exception as e:
            logger.error(f"Failed to create schedule: {e}")
            return json.dumps({"error": str(e)})

    async def alist_schedules(self, enabled_only: bool = False) -> str:
        """List all existing schedules.

        Args:
            enabled_only (bool): If True, only return enabled schedules. Defaults to False.

        Returns:
            str: JSON string with the list of schedules.
        """
        try:
            enabled_filter = True if enabled_only else None
            schedules = await self.manager.alist(enabled=enabled_filter)
            result = [
                {
                    "id": s.id,
                    "name": s.name,
                    "cron": s.cron_expr,
                    "endpoint": s.endpoint,
                    "timezone": s.timezone,
                    "enabled": s.enabled,
                    "description": s.description,
                }
                for s in schedules
            ]
            return json.dumps({"schedules": result, "count": len(result)})
        except Exception as e:
            logger.error(f"Failed to list schedules: {e}")
            return json.dumps({"error": str(e)})

    async def aget_schedule(self, schedule_id: str) -> str:
        """Get details of a specific schedule by its ID.

        Args:
            schedule_id (str): The ID of the schedule to retrieve.

        Returns:
            str: JSON string with the schedule details.
        """
        try:
            schedule = await self.manager.aget(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "id": schedule.id,
                    "name": schedule.name,
                    "cron": schedule.cron_expr,
                    "endpoint": schedule.endpoint,
                    "method": schedule.method,
                    "timezone": schedule.timezone,
                    "enabled": schedule.enabled,
                    "description": schedule.description,
                    "payload": schedule.payload,
                }
            )
        except Exception as e:
            logger.error(f"Failed to get schedule: {e}")
            return json.dumps({"error": str(e)})

    async def adelete_schedule(self, schedule_id: str) -> str:
        """Delete a schedule by its ID. This permanently removes the schedule.

        Args:
            schedule_id (str): The ID of the schedule to delete.

        Returns:
            str: JSON string confirming deletion.
        """
        try:
            deleted = await self.manager.adelete(schedule_id)
            if deleted:
                return json.dumps({"status": "deleted", "id": schedule_id})
            return json.dumps({"error": f"Schedule not found or could not be deleted: {schedule_id}"})
        except Exception as e:
            logger.error(f"Failed to delete schedule: {e}")
            return json.dumps({"error": str(e)})

    async def aenable_schedule(self, schedule_id: str) -> str:
        """Enable a disabled schedule so it starts running again.

        Args:
            schedule_id (str): The ID of the schedule to enable.

        Returns:
            str: JSON string with the updated schedule details.
        """
        try:
            schedule = await self.manager.aenable(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "status": "enabled",
                    "id": schedule.id,
                    "name": schedule.name,
                    "enabled": schedule.enabled,
                }
            )
        except Exception as e:
            logger.error(f"Failed to enable schedule: {e}")
            return json.dumps({"error": str(e)})

    async def adisable_schedule(self, schedule_id: str) -> str:
        """Disable a schedule so it stops running. Can be re-enabled later.

        Args:
            schedule_id (str): The ID of the schedule to disable.

        Returns:
            str: JSON string with the updated schedule details.
        """
        try:
            schedule = await self.manager.adisable(schedule_id)
            if schedule is None:
                return json.dumps({"error": f"Schedule not found: {schedule_id}"})
            return json.dumps(
                {
                    "status": "disabled",
                    "id": schedule.id,
                    "name": schedule.name,
                    "enabled": schedule.enabled,
                }
            )
        except Exception as e:
            logger.error(f"Failed to disable schedule: {e}")
            return json.dumps({"error": str(e)})

    async def aget_schedule_runs(self, schedule_id: str, limit: int = 10) -> str:
        """Get the run history for a schedule.

        Args:
            schedule_id (str): The ID of the schedule to get runs for.
            limit (int): Maximum number of runs to return. Defaults to 10.

        Returns:
            str: JSON string with the list of runs.
        """
        try:
            runs = await self.manager.aget_runs(schedule_id, limit=limit)
            result = [
                {
                    "id": r.id,
                    "status": r.status,
                    "triggered_at": r.triggered_at,
                    "completed_at": r.completed_at,
                    "error": r.error,
                }
                for r in runs
            ]
            return json.dumps({"runs": result, "count": len(result)})
        except Exception as e:
            logger.error(f"Failed to get schedule runs: {e}")
            return json.dumps({"error": str(e)})



"""
ReasoningManager - Centralized manager for all reasoning operations.

This module consolidates reasoning logic from the Agent class into a single,
maintainable manager that handles:
- Native reasoning models (DeepSeek, Anthropic, OpenAI, Gemini, etc.)
- Default Chain-of-Thought reasoning
- Both streaming and non-streaming modes
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

from agno.models.base import Model
from agno.models.message import Message
from agno.reasoning.step import NextAction, ReasoningStep, ReasoningSteps
from agno.run.base import RunContext
from agno.run.messages import RunMessages
from agno.tools import Toolkit
from agno.tools.function import Function
from agno.utils.log import log_debug, log_error, log_info, log_warning

if TYPE_CHECKING:
    from agno.agent import Agent
    from agno.metrics import RunMetrics
    from agno.run.agent import RunOutput


class ReasoningEventType(str, Enum):
    """Types of reasoning events that can be emitted."""

    started = "reasoning_started"
    content_delta = "reasoning_content_delta"
    step = "reasoning_step"
    completed = "reasoning_completed"
    error = "reasoning_error"


@dataclass
class ReasoningEvent:
    """
    A unified reasoning event that can be converted to Agent or Team specific events.

    This allows the ReasoningManager to emit events without knowing about the
    specific event types used by Agent or Team.
    """

    event_type: ReasoningEventType
    # For content_delta events
    reasoning_content: Optional[str] = None
    # For step events
    reasoning_step: Optional[ReasoningStep] = None
    # For completed events
    reasoning_steps: List[ReasoningStep] = field(default_factory=list)
    # For error events
    error: Optional[str] = None
    # The message to append to run_messages (for native reasoning)
    message: Optional[Message] = None
    # All reasoning messages (for updating run_output)
    reasoning_messages: List[Message] = field(default_factory=list)


@dataclass
class ReasoningConfig:
    """Configuration for reasoning operations."""

    reasoning_model: Optional[Model] = None
    reasoning_agent: Optional["Agent"] = None
    min_steps: int = 1
    max_steps: int = 10
    tools: Optional[List[Union[Toolkit, Callable, Function, Dict]]] = None
    tool_call_limit: Optional[int] = None
    use_json_mode: bool = False
    telemetry: bool = True
    debug_mode: bool = False
    debug_level: Literal[1, 2] = 1
    run_context: Optional[RunContext] = None
    run_metrics: Optional["RunMetrics"] = None


@dataclass
class ReasoningResult:
    """Result from a reasoning operation."""

    message: Optional[Message] = None
    steps: List[ReasoningStep] = field(default_factory=list)
    reasoning_messages: List[Message] = field(default_factory=list)
    success: bool = True
    error: Optional[str] = None


class ReasoningManager:
    """
    Centralized manager for all reasoning operations.

    Handles both native reasoning models (DeepSeek, Anthropic, OpenAI, etc.)
    and default Chain-of-Thought reasoning with a clean, unified interface.
    """

    def __init__(self, config: ReasoningConfig):
        self.config = config
        self._reasoning_agent: Optional["Agent"] = None
        self._model_type: Optional[str] = None

    @property
    def reasoning_model(self) -> Optional[Model]:
        return self.config.reasoning_model

    def _detect_model_type(self, model: Model) -> Optional[str]:
        """Detect the type of reasoning model."""
        from agno.reasoning.anthropic import is_anthropic_reasoning_model
        from agno.reasoning.azure_ai_foundry import is_ai_foundry_reasoning_model
        from agno.reasoning.deepseek import is_deepseek_reasoning_model
        from agno.reasoning.gemini import is_gemini_reasoning_model
        from agno.reasoning.groq import is_groq_reasoning_model
        from agno.reasoning.ollama import is_ollama_reasoning_model
        from agno.reasoning.openai import is_openai_reasoning_model
        from agno.reasoning.vertexai import is_vertexai_reasoning_model

        if is_deepseek_reasoning_model(model):
            return "deepseek"
        if is_anthropic_reasoning_model(model):
            return "anthropic"
        if is_openai_reasoning_model(model):
            return "openai"
        if is_groq_reasoning_model(model):
            return "groq"
        if is_ollama_reasoning_model(model):
            return "ollama"
        if is_ai_foundry_reasoning_model(model):
            return "ai_foundry"
        if is_gemini_reasoning_model(model):
            return "gemini"
        if is_vertexai_reasoning_model(model):
            return "vertexai"
        return None

    def _get_reasoning_agent(self, model: Model) -> "Agent":
        """Get or create a reasoning agent for the given model."""
        if self.config.reasoning_agent is not None:
            return self.config.reasoning_agent

        from agno.reasoning.helpers import get_reasoning_agent

        return get_reasoning_agent(
            reasoning_model=model,
            telemetry=self.config.telemetry,
            debug_mode=self.config.debug_mode,
            debug_level=self.config.debug_level,
            run_context=self.config.run_context,
        )

    def _get_default_reasoning_agent(self, model: Model) -> Optional["Agent"]:
        """Get or create a default CoT reasoning agent."""
        if self.config.reasoning_agent is not None:
            return self.config.reasoning_agent

        from agno.reasoning.default import get_default_reasoning_agent

        return get_default_reasoning_agent(
            reasoning_model=model,
            min_steps=self.config.min_steps,
            max_steps=self.config.max_steps,
            tools=self.config.tools,
            tool_call_limit=self.config.tool_call_limit,
            use_json_mode=self.config.use_json_mode,
            telemetry=self.config.telemetry,
            debug_mode=self.config.debug_mode,
            debug_level=self.config.debug_level,
            run_context=self.config.run_context,
        )

    def is_native_reasoning_model(self, model: Optional[Model] = None) -> bool:
        """Check if the model is a native reasoning model."""
        model = model or self.config.reasoning_model
        if model is None:
            return False
        return self._detect_model_type(model) is not None

    # =========================================================================
    # Native Model Reasoning (Non-Streaming)
    # =========================================================================

    def get_native_reasoning(self, model: Model, messages: List[Message]) -> ReasoningResult:
        """Get reasoning from a native reasoning model (non-streaming)."""
        model_type = self._detect_model_type(model)
        if model_type is None:
            return ReasoningResult(success=False, error="Not a native reasoning model")

        reasoning_agent = self._get_reasoning_agent(model)
        reasoning_message: Optional[Message] = None
        run_metrics = self.config.run_metrics

        try:
            if model_type == "deepseek":
                from agno.reasoning.deepseek import get_deepseek_reasoning

                log_debug("Starting DeepSeek Reasoning", center=True, symbol="=")
                reasoning_message = get_deepseek_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "anthropic":
                from agno.reasoning.anthropic import get_anthropic_reasoning

                log_debug("Starting Anthropic Claude Reasoning", center=True, symbol="=")
                reasoning_message = get_anthropic_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "openai":
                from agno.reasoning.openai import get_openai_reasoning

                log_debug("Starting OpenAI Reasoning", center=True, symbol="=")
                reasoning_message = get_openai_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "groq":
                from agno.reasoning.groq import get_groq_reasoning

                log_debug("Starting Groq Reasoning", center=True, symbol="=")
                reasoning_message = get_groq_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "ollama":
                from agno.reasoning.ollama import get_ollama_reasoning

                log_debug("Starting Ollama Reasoning", center=True, symbol="=")
                reasoning_message = get_ollama_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "ai_foundry":
                from agno.reasoning.azure_ai_foundry import get_ai_foundry_reasoning

                log_debug("Starting Azure AI Foundry Reasoning", center=True, symbol="=")
                reasoning_message = get_ai_foundry_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "gemini":
                from agno.reasoning.gemini import get_gemini_reasoning

                log_debug("Starting Gemini Reasoning", center=True, symbol="=")
                reasoning_message = get_gemini_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "vertexai":
                from agno.reasoning.vertexai import get_vertexai_reasoning

                log_debug("Starting VertexAI Reasoning", center=True, symbol="=")
                reasoning_message = get_vertexai_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

        except Exception as e:
            log_error(f"Reasoning error: {e}")
            return ReasoningResult(success=False, error=str(e))

        if reasoning_message is None:
            return ReasoningResult(
                success=False,
                error="Reasoning response is None",
            )

        return ReasoningResult(
            message=reasoning_message,
            steps=[ReasoningStep(result=reasoning_message.content)],
            reasoning_messages=[reasoning_message],
            success=True,
        )

    async def aget_native_reasoning(self, model: Model, messages: List[Message]) -> ReasoningResult:
        """Get reasoning from a native reasoning model asynchronously (non-streaming)."""
        model_type = self._detect_model_type(model)
        if model_type is None:
            return ReasoningResult(success=False, error="Not a native reasoning model")

        reasoning_agent = self._get_reasoning_agent(model)
        reasoning_message: Optional[Message] = None
        run_metrics = self.config.run_metrics

        try:
            if model_type == "deepseek":
                from agno.reasoning.deepseek import aget_deepseek_reasoning

                log_debug("Starting DeepSeek Reasoning", center=True, symbol="=")
                reasoning_message = await aget_deepseek_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "anthropic":
                from agno.reasoning.anthropic import aget_anthropic_reasoning

                log_debug("Starting Anthropic Claude Reasoning", center=True, symbol="=")
                reasoning_message = await aget_anthropic_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "openai":
                from agno.reasoning.openai import aget_openai_reasoning

                log_debug("Starting OpenAI Reasoning", center=True, symbol="=")
                reasoning_message = await aget_openai_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "groq":
                from agno.reasoning.groq import aget_groq_reasoning

                log_debug("Starting Groq Reasoning", center=True, symbol="=")
                reasoning_message = await aget_groq_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "ollama":
                from agno.reasoning.ollama import aget_ollama_reasoning

                log_debug("Starting Ollama Reasoning", center=True, symbol="=")
                reasoning_message = await aget_ollama_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "ai_foundry":
                from agno.reasoning.azure_ai_foundry import aget_ai_foundry_reasoning

                log_debug("Starting Azure AI Foundry Reasoning", center=True, symbol="=")
                reasoning_message = await aget_ai_foundry_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "gemini":
                from agno.reasoning.gemini import aget_gemini_reasoning

                log_debug("Starting Gemini Reasoning", center=True, symbol="=")
                reasoning_message = await aget_gemini_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

            elif model_type == "vertexai":
                from agno.reasoning.vertexai import aget_vertexai_reasoning

                log_debug("Starting VertexAI Reasoning", center=True, symbol="=")
                reasoning_message = await aget_vertexai_reasoning(reasoning_agent, messages, run_metrics=run_metrics)

        except Exception as e:
            log_error(f"Reasoning error: {e}")
            return ReasoningResult(success=False, error=str(e))

        if reasoning_message is None:
            return ReasoningResult(
                success=False,
                error="Reasoning response is None",
            )

        return ReasoningResult(
            message=reasoning_message,
            steps=[ReasoningStep(result=reasoning_message.content)],
            reasoning_messages=[reasoning_message],
            success=True,
        )

    # =========================================================================
    # Native Model Reasoning (Streaming)
    # =========================================================================

    def stream_native_reasoning(
        self, model: Model, messages: List[Message]
    ) -> Iterator[Tuple[Optional[str], Optional[ReasoningResult]]]:
        """
        Stream reasoning from a native reasoning model.

        Yields:
            Tuple of (reasoning_content_delta, final_result)
            - During streaming: (reasoning_content_delta, None)
            - At the end: (None, ReasoningResult)
        """
        model_type = self._detect_model_type(model)
        if model_type is None:
            yield (None, ReasoningResult(success=False, error="Not a native reasoning model"))
            return

        reasoning_agent = self._get_reasoning_agent(model)

        # Currently only DeepSeek and Anthropic support streaming
        if model_type == "deepseek":
            from agno.reasoning.deepseek import get_deepseek_reasoning_stream

            log_debug("Starting DeepSeek Reasoning (streaming)", center=True, symbol="=")
            final_message: Optional[Message] = None
            for reasoning_delta, message in get_deepseek_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "anthropic":
            from agno.reasoning.anthropic import get_anthropic_reasoning_stream

            log_debug("Starting Anthropic Claude Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_anthropic_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "gemini":
            from agno.reasoning.gemini import get_gemini_reasoning_stream

            log_debug("Starting Gemini Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_gemini_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "openai":
            from agno.reasoning.openai import get_openai_reasoning_stream

            log_debug("Starting OpenAI Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_openai_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "vertexai":
            from agno.reasoning.vertexai import get_vertexai_reasoning_stream

            log_debug("Starting VertexAI Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_vertexai_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "ai_foundry":
            from agno.reasoning.azure_ai_foundry import get_ai_foundry_reasoning_stream

            log_debug("Starting Azure AI Foundry Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_ai_foundry_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "groq":
            from agno.reasoning.groq import get_groq_reasoning_stream

            log_debug("Starting Groq Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_groq_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "ollama":
            from agno.reasoning.ollama import get_ollama_reasoning_stream

            log_debug("Starting Ollama Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            for reasoning_delta, message in get_ollama_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        else:
            # Fall back to non-streaming for other models
            result = self.get_native_reasoning(model, messages)
            yield (None, result)

    async def astream_native_reasoning(
        self, model: Model, messages: List[Message]
    ) -> AsyncIterator[Tuple[Optional[str], Optional[ReasoningResult]]]:
        """
        Stream reasoning from a native reasoning model asynchronously.

        Yields:
            Tuple of (reasoning_content_delta, final_result)
            - During streaming: (reasoning_content_delta, None)
            - At the end: (None, ReasoningResult)
        """
        model_type = self._detect_model_type(model)
        if model_type is None:
            yield (None, ReasoningResult(success=False, error="Not a native reasoning model"))
            return

        reasoning_agent = self._get_reasoning_agent(model)

        # Currently only DeepSeek and Anthropic support streaming
        if model_type == "deepseek":
            from agno.reasoning.deepseek import aget_deepseek_reasoning_stream

            log_debug("Starting DeepSeek Reasoning (streaming)", center=True, symbol="=")
            final_message: Optional[Message] = None
            async for reasoning_delta, message in aget_deepseek_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "anthropic":
            from agno.reasoning.anthropic import aget_anthropic_reasoning_stream

            log_debug("Starting Anthropic Claude Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_anthropic_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "gemini":
            from agno.reasoning.gemini import aget_gemini_reasoning_stream

            log_debug("Starting Gemini Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_gemini_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "openai":
            from agno.reasoning.openai import aget_openai_reasoning_stream

            log_debug("Starting OpenAI Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_openai_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "vertexai":
            from agno.reasoning.vertexai import aget_vertexai_reasoning_stream

            log_debug("Starting VertexAI Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_vertexai_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "ai_foundry":
            from agno.reasoning.azure_ai_foundry import aget_ai_foundry_reasoning_stream

            log_debug("Starting Azure AI Foundry Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_ai_foundry_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "groq":
            from agno.reasoning.groq import aget_groq_reasoning_stream

            log_debug("Starting Groq Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_groq_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        elif model_type == "ollama":
            from agno.reasoning.ollama import aget_ollama_reasoning_stream

            log_debug("Starting Ollama Reasoning (streaming)", center=True, symbol="=")
            final_message = None
            async for reasoning_delta, message in aget_ollama_reasoning_stream(reasoning_agent, messages):
                if reasoning_delta is not None:
                    yield (reasoning_delta, None)
                if message is not None:
                    final_message = message

            if final_message:
                yield (
                    None,
                    ReasoningResult(
                        message=final_message,
                        steps=[ReasoningStep(result=final_message.content)],
                        reasoning_messages=[final_message],
                        success=True,
                    ),
                )
            else:
                yield (None, ReasoningResult(success=False, error="No reasoning content"))

        else:
            # Fall back to non-streaming for other models
            result = await self.aget_native_reasoning(model, messages)
            yield (None, result)

    # =========================================================================
    # Default Chain-of-Thought Reasoning
    # =========================================================================

    def run_default_reasoning(
        self, model: Model, run_messages: RunMessages
    ) -> Iterator[Tuple[Optional[ReasoningStep], Optional[ReasoningResult]]]:
        """
        Run default Chain-of-Thought reasoning.

        Yields:
            Tuple of (reasoning_step, final_result)
            - During reasoning: (ReasoningStep, None)
            - At the end: (None, ReasoningResult)
        """
        from agno.reasoning.helpers import get_next_action, update_messages_with_reasoning

        reasoning_agent = self._get_default_reasoning_agent(model)
        if reasoning_agent is None:
            yield (None, ReasoningResult(success=False, error="Reasoning agent is None"))
            return

        # Validate reasoning agent output schema
        if (
            reasoning_agent.output_schema is not None
            and isinstance(reasoning_agent.output_schema, type)
            and not issubclass(reasoning_agent.output_schema, ReasoningSteps)
        ):
            yield (
                None,
                ReasoningResult(
                    success=False,
                    error="Reasoning agent response model should be ReasoningSteps",
                ),
            )
            return

        step_count = 1
        next_action = NextAction.CONTINUE
        reasoning_messages: List[Message] = []
        all_reasoning_steps: List[ReasoningStep] = []

        log_debug("Starting Reasoning", center=True, symbol="=")

        while next_action == NextAction.CONTINUE and step_count < self.config.max_steps:
            log_debug(f"Step {step_count}", center=True, symbol="=")
            try:
                reasoning_agent_response: RunOutput = reasoning_agent.run(input=run_messages.get_input_messages())

                # Accumulate reasoning model metrics
                if self.config.run_metrics is not None:
                    from agno.metrics import accumulate_eval_metrics

                    accumulate_eval_metrics(
                        reasoning_agent_response.metrics, self.config.run_metrics, prefix="reasoning"
                    )

                if reasoning_agent_response.content is None or reasoning_agent_response.messages is None:
                    log_warning("Reasoning error. Reasoning response is empty")
                    break

                if isinstance(reasoning_agent_response.content, str):
                    log_warning("Reasoning error. Content is a string, not structured output")
                    break

                if (
                    reasoning_agent_response.content.reasoning_steps is None
                    or len(reasoning_agent_response.content.reasoning_steps) == 0
                ):
                    log_warning("Reasoning error. Reasoning steps are empty")
                    break

                reasoning_steps: List[ReasoningStep] = reasoning_agent_response.content.reasoning_steps
                all_reasoning_steps.extend(reasoning_steps)

                # Yield each reasoning step
                for step in reasoning_steps:
                    yield (step, None)

                # Extract reasoning messages
                first_assistant_index = next(
                    (i for i, m in enumerate(reasoning_agent_response.messages) if m.role == "assistant"),
                    len(reasoning_agent_response.messages),
                )
                reasoning_messages = reasoning_agent_response.messages[first_assistant_index:]

                # Get the next action
                next_action = get_next_action(reasoning_steps[-1])
                if next_action == NextAction.FINAL_ANSWER:
                    break

            except Exception as e:
                log_error(f"Reasoning error: {e}")
                break

            step_count += 1

        log_debug(f"Total Reasoning steps: {len(all_reasoning_steps)}")
        log_debug("Reasoning finished", center=True, symbol="=")

        # Update messages with reasoning
        update_messages_with_reasoning(
            run_messages=run_messages,
            reasoning_messages=reasoning_messages,
        )

        # Yield final result
        yield (
            None,
            ReasoningResult(
                steps=all_reasoning_steps,
                reasoning_messages=reasoning_messages,
                success=True,
            ),
        )

    async def arun_default_reasoning(
        self, model: Model, run_messages: RunMessages
    ) -> AsyncIterator[Tuple[Optional[ReasoningStep], Optional[ReasoningResult]]]:
        """
        Run default Chain-of-Thought reasoning asynchronously.

        Yields:
            Tuple of (reasoning_step, final_result)
            - During reasoning: (ReasoningStep, None)
            - At the end: (None, ReasoningResult)
        """
        from agno.reasoning.helpers import get_next_action, update_messages_with_reasoning

        reasoning_agent = self._get_default_reasoning_agent(model)
        if reasoning_agent is None:
            yield (None, ReasoningResult(success=False, error="Reasoning agent is None"))
            return

        # Validate reasoning agent output schema
        if (
            reasoning_agent.output_schema is not None
            and isinstance(reasoning_agent.output_schema, type)
            and not issubclass(reasoning_agent.output_schema, ReasoningSteps)
        ):
            yield (
                None,
                ReasoningResult(
                    success=False,
                    error="Reasoning agent response model should be ReasoningSteps",
                ),
            )
            return

        step_count = 1
        next_action = NextAction.CONTINUE
        reasoning_messages: List[Message] = []
        all_reasoning_steps: List[ReasoningStep] = []

        log_debug("Starting Reasoning", center=True, symbol="=")

        while next_action == NextAction.CONTINUE and step_count < self.config.max_steps:
            log_debug(f"Step {step_count}", center=True, symbol="=")
            step_count += 1
            try:
                reasoning_agent_response: RunOutput = await reasoning_agent.arun(  # type: ignore[misc]
                    input=run_messages.get_input_messages()
                )

                # Accumulate reasoning model metrics
                if self.config.run_metrics is not None:
                    from agno.metrics import accumulate_eval_metrics

                    accumulate_eval_metrics(
                        reasoning_agent_response.metrics, self.config.run_metrics, prefix="reasoning"
                    )

                if reasoning_agent_response.content is None or reasoning_agent_response.messages is None:
                    log_warning("Reasoning error. Reasoning response is empty")
                    break

                if isinstance(reasoning_agent_response.content, str):
                    log_warning("Reasoning error. Content is a string, not structured output")
                    break

                if reasoning_agent_response.content.reasoning_steps is None:
                    log_warning("Reasoning error. Reasoning steps are empty")
                    break

                reasoning_steps: List[ReasoningStep] = reasoning_agent_response.content.reasoning_steps
                all_reasoning_steps.extend(reasoning_steps)

                # Yield each reasoning step
                for step in reasoning_steps:
                    yield (step, None)

                # Extract reasoning messages
                first_assistant_index = next(
                    (i for i, m in enumerate(reasoning_agent_response.messages) if m.role == "assistant"),
                    len(reasoning_agent_response.messages),
                )
                reasoning_messages = reasoning_agent_response.messages[first_assistant_index:]

                # Get the next action
                next_action = get_next_action(reasoning_steps[-1])
                if next_action == NextAction.FINAL_ANSWER:
                    break

            except Exception as e:
                log_error(f"Reasoning error: {e}")
                break

        log_debug(f"Total Reasoning steps: {len(all_reasoning_steps)}")
        log_debug("Reasoning finished", center=True, symbol="=")

        # Update messages with reasoning
        update_messages_with_reasoning(
            run_messages=run_messages,
            reasoning_messages=reasoning_messages,
        )

        # Yield final result
        yield (
            None,
            ReasoningResult(
                steps=all_reasoning_steps,
                reasoning_messages=reasoning_messages,
                success=True,
            ),
        )

    def reason(
        self,
        run_messages: RunMessages,
        stream: bool = False,
    ) -> Iterator[ReasoningEvent]:
        """
        Run reasoning and yield ReasoningEvent objects.

        Args:
            run_messages: The messages to reason about
            stream: Whether to stream reasoning content

        Yields:
            ReasoningEvent objects for each stage of reasoning
        """
        # Get the reasoning model
        reasoning_model: Optional[Model] = self.config.reasoning_model
        reasoning_model_provided = reasoning_model is not None

        if reasoning_model is None:
            yield ReasoningEvent(
                event_type=ReasoningEventType.error,
                error="Reasoning model is None",
            )
            return

        # Yield started event
        yield ReasoningEvent(event_type=ReasoningEventType.started)

        # Check if this is a native reasoning model
        if reasoning_model_provided and self.is_native_reasoning_model(reasoning_model):
            # Use streaming for native models when stream is enabled
            if stream:
                yield from self._stream_native_reasoning_events(reasoning_model, run_messages)
            else:
                yield from self._get_native_reasoning_events(reasoning_model, run_messages)
        else:
            # Use default Chain-of-Thought reasoning
            if reasoning_model_provided:
                log_info(
                    f"Reasoning model: {reasoning_model.__class__.__name__} is not a native reasoning model, "
                    "defaulting to manual Chain-of-Thought reasoning"
                )
            yield from self._run_default_reasoning_events(reasoning_model, run_messages)

    async def areason(
        self,
        run_messages: RunMessages,
        stream: bool = False,
    ) -> AsyncIterator[ReasoningEvent]:
        """
        Unified async reasoning interface that yields ReasoningEvent objects.

        This method handles all reasoning logic and yields events that can be
        converted to Agent or Team specific events by the caller.

        Args:
            run_messages: The messages to reason about
            stream: Whether to stream reasoning content deltas

        Yields:
            ReasoningEvent objects for each stage of reasoning
        """
        # Get the reasoning model
        reasoning_model: Optional[Model] = self.config.reasoning_model
        reasoning_model_provided = reasoning_model is not None

        if reasoning_model is None:
            yield ReasoningEvent(
                event_type=ReasoningEventType.error,
                error="Reasoning model is None",
            )
            return

        # Yield started event
        yield ReasoningEvent(event_type=ReasoningEventType.started)

        # Check if this is a native reasoning model
        if reasoning_model_provided and self.is_native_reasoning_model(reasoning_model):
            # Use streaming for native models when stream is enabled
            if stream:
                async for event in self._astream_native_reasoning_events(reasoning_model, run_messages):
                    yield event
            else:
                async for event in self._aget_native_reasoning_events(reasoning_model, run_messages):
                    yield event
        else:
            # Use default Chain-of-Thought reasoning
            if reasoning_model_provided:
                log_info(
                    f"Reasoning model: {reasoning_model.__class__.__name__} is not a native reasoning model, "
                    "defaulting to manual Chain-of-Thought reasoning"
                )
            async for event in self._arun_default_reasoning_events(reasoning_model, run_messages):
                yield event

    def _stream_native_reasoning_events(self, model: Model, run_messages: RunMessages) -> Iterator[ReasoningEvent]:
        """Stream native reasoning and yield ReasoningEvent objects."""
        messages = run_messages.get_input_messages()

        for reasoning_delta, result in self.stream_native_reasoning(model, messages):
            if reasoning_delta is not None:
                yield ReasoningEvent(
                    event_type=ReasoningEventType.content_delta,
                    reasoning_content=reasoning_delta,
                )
            if result is not None:
                if not result.success:
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.error,
                        error=result.error,
                    )
                    return
                if result.message:
                    run_messages.messages.append(result.message)
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.completed,
                        reasoning_steps=result.steps,
                        message=result.message,
                        reasoning_messages=result.reasoning_messages,
                    )

    def _get_native_reasoning_events(self, model: Model, run_messages: RunMessages) -> Iterator[ReasoningEvent]:
        """Get native reasoning (non-streaming) and yield ReasoningEvent objects."""
        messages = run_messages.get_input_messages()
        result = self.get_native_reasoning(model, messages)

        if not result.success:
            yield ReasoningEvent(
                event_type=ReasoningEventType.error,
                error=result.error,
            )
            return

        if result.message:
            run_messages.messages.append(result.message)
            yield ReasoningEvent(
                event_type=ReasoningEventType.completed,
                reasoning_steps=result.steps,
                message=result.message,
                reasoning_messages=result.reasoning_messages,
            )

    def _run_default_reasoning_events(self, model: Model, run_messages: RunMessages) -> Iterator[ReasoningEvent]:
        """Run default CoT reasoning and yield ReasoningEvent objects."""
        all_reasoning_steps: List[ReasoningStep] = []

        for reasoning_step, result in self.run_default_reasoning(model, run_messages):
            if reasoning_step is not None:
                all_reasoning_steps.append(reasoning_step)
                yield ReasoningEvent(
                    event_type=ReasoningEventType.step,
                    reasoning_step=reasoning_step,
                )
            if result is not None:
                if not result.success:
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.error,
                        error=result.error,
                    )
                    return

        # Yield completed event with all steps
        if all_reasoning_steps:
            yield ReasoningEvent(
                event_type=ReasoningEventType.completed,
                reasoning_steps=all_reasoning_steps,
            )

    async def _astream_native_reasoning_events(
        self, model: Model, run_messages: RunMessages
    ) -> AsyncIterator[ReasoningEvent]:
        """Stream native reasoning asynchronously and yield ReasoningEvent objects."""
        messages = run_messages.get_input_messages()

        async for reasoning_delta, result in self.astream_native_reasoning(model, messages):
            if reasoning_delta is not None:
                yield ReasoningEvent(
                    event_type=ReasoningEventType.content_delta,
                    reasoning_content=reasoning_delta,
                )
            if result is not None:
                if not result.success:
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.error,
                        error=result.error,
                    )
                    return
                if result.message:
                    run_messages.messages.append(result.message)
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.completed,
                        reasoning_steps=result.steps,
                        message=result.message,
                        reasoning_messages=result.reasoning_messages,
                    )

    async def _aget_native_reasoning_events(
        self, model: Model, run_messages: RunMessages
    ) -> AsyncIterator[ReasoningEvent]:
        """Get native reasoning asynchronously (non-streaming) and yield ReasoningEvent objects."""
        messages = run_messages.get_input_messages()
        result = await self.aget_native_reasoning(model, messages)

        if not result.success:
            yield ReasoningEvent(
                event_type=ReasoningEventType.error,
                error=result.error,
            )
            return

        if result.message:
            run_messages.messages.append(result.message)
            yield ReasoningEvent(
                event_type=ReasoningEventType.completed,
                reasoning_steps=result.steps,
                message=result.message,
                reasoning_messages=result.reasoning_messages,
            )

    async def _arun_default_reasoning_events(
        self, model: Model, run_messages: RunMessages
    ) -> AsyncIterator[ReasoningEvent]:
        """Run default CoT reasoning asynchronously and yield ReasoningEvent objects."""
        all_reasoning_steps: List[ReasoningStep] = []

        async for reasoning_step, result in self.arun_default_reasoning(model, run_messages):
            if reasoning_step is not None:
                all_reasoning_steps.append(reasoning_step)
                yield ReasoningEvent(
                    event_type=ReasoningEventType.step,
                    reasoning_step=reasoning_step,
                )
            if result is not None:
                if not result.success:
                    yield ReasoningEvent(
                        event_type=ReasoningEventType.error,
                        error=result.error,
                    )
                    return

        # Yield completed event with all steps
        if all_reasoning_steps:
            yield ReasoningEvent(
                event_type=ReasoningEventType.completed,
                reasoning_steps=all_reasoning_steps,
            )



"""GCS content loader for Knowledge.

Provides methods for loading content from Google Cloud Storage.
"""

# mypy: disable-error-code="attr-defined"

from io import BytesIO
from typing import Any, Dict, Optional, cast

from agno.knowledge.content import Content, ContentStatus
from agno.knowledge.loaders.base import BaseLoader
from agno.knowledge.reader import Reader
from agno.knowledge.remote_content.base import BaseStorageConfig
from agno.knowledge.remote_content.gcs import GcsConfig
from agno.knowledge.remote_content.remote_content import GCSContent
from agno.utils.log import log_info, log_warning
from agno.utils.string import generate_id


class GCSLoader(BaseLoader):
    """Loader for Google Cloud Storage content."""

    # ==========================================
    # GCS HELPERS (shared between sync/async)
    # ==========================================

    def _validate_gcs_config(
        self,
        content: Content,
        config: Optional[BaseStorageConfig],
    ) -> Optional[GcsConfig]:
        """Validate and extract GCS config.

        Returns:
            GcsConfig if valid, None otherwise (GCS can work without explicit config)
        """
        return cast(GcsConfig, config) if isinstance(config, GcsConfig) else None

    def _get_gcs_client(self, gcs_config: Optional[GcsConfig]):
        """Get a GCS client.

        Requires the `google-cloud-storage` package.
        """
        try:
            from google.cloud import storage  # type: ignore
        except ImportError:
            raise ImportError(
                "The `google-cloud-storage` package is not installed. "
                "Please install it via `pip install google-cloud-storage`."
            )

        if gcs_config and gcs_config.credentials_path:
            return storage.Client.from_service_account_json(gcs_config.credentials_path)
        elif gcs_config and gcs_config.project:
            return storage.Client(project=gcs_config.project)
        else:
            return storage.Client()

    def _build_gcs_metadata(
        self,
        gcs_config: Optional[GcsConfig],
        bucket_name: str,
        blob_name: str,
    ) -> Dict[str, str]:
        """Build GCS-specific metadata dictionary."""
        metadata: Dict[str, str] = {
            "source_type": "gcs",
            "gcs_bucket": bucket_name,
            "gcs_blob_name": blob_name,
        }
        if gcs_config:
            metadata["source_config_id"] = gcs_config.id
            metadata["source_config_name"] = gcs_config.name
        return metadata

    def _build_gcs_virtual_path(self, bucket_name: str, blob_name: str) -> str:
        """Build virtual path for GCS content."""
        return f"gcs://{bucket_name}/{blob_name}"

    # ==========================================
    # GCS LOADERS
    # ==========================================

    async def _aload_from_gcs(
        self,
        content: Content,
        upsert: bool,
        skip_if_exists: bool,
        config: Optional[BaseStorageConfig] = None,
    ):
        """Load content from Google Cloud Storage (async).

        Note: Uses sync google-cloud-storage calls as it doesn't have an async API.
        """
        try:
            from google.cloud import storage  # type: ignore  # noqa: F401
        except ImportError:
            raise ImportError(
                "The `google-cloud-storage` package is not installed. "
                "Please install it via `pip install google-cloud-storage`."
            )

        log_warning(
            "GCS content loading has limited features. "
            "Recursive folder traversal, rich metadata, and improved naming are coming in a future release."
        )

        remote_content: GCSContent = cast(GCSContent, content.remote_content)
        gcs_config = self._validate_gcs_config(content, config)

        # Get or create bucket
        bucket = remote_content.bucket
        if bucket is None and remote_content.bucket_name:
            client = self._get_gcs_client(gcs_config)
            bucket = client.bucket(remote_content.bucket_name)

        # Identify objects to read
        objects_to_read = []
        if remote_content.blob_name is not None:
            objects_to_read.append(bucket.blob(remote_content.blob_name))  # type: ignore
        elif remote_content.prefix is not None:
            objects_to_read.extend(bucket.list_blobs(prefix=remote_content.prefix))  # type: ignore
        else:
            objects_to_read.extend(bucket.list_blobs())  # type: ignore

        if objects_to_read:
            log_info(f"Processing {len(objects_to_read)} file(s) from GCS")

        bucket_name = remote_content.bucket_name or (bucket.name if bucket else "unknown")
        is_folder_upload = len(objects_to_read) > 1
        root_path = remote_content.prefix or ""

        for gcs_object in objects_to_read:
            blob_name = gcs_object.name
            file_name = blob_name.split("/")[-1]

            # Build metadata and virtual path using helpers
            virtual_path = self._build_gcs_virtual_path(bucket_name, blob_name)
            gcs_metadata = self._build_gcs_metadata(gcs_config, bucket_name, blob_name)
            merged_metadata: Dict[str, Any] = self._merge_metadata(gcs_metadata, content.metadata)

            # Compute content name using base helper
            content_name = self._compute_content_name(blob_name, file_name, content.name, root_path, is_folder_upload)

            # Create content entry
            content_entry = Content(
                name=content_name,
                description=content.description,
                path=virtual_path,
                status=ContentStatus.PROCESSING,
                metadata=merged_metadata,
                file_type="gcs",
            )
            content_entry.content_hash = self._build_content_hash(content_entry)
            content_entry.id = generate_id(content_entry.content_hash)

            await self._ainsert_contents_db(content_entry)

            if self._should_skip(content_entry.content_hash, skip_if_exists):
                content_entry.status = ContentStatus.COMPLETED
                await self._aupdate_content(content_entry)
                continue

            # Select reader
            reader = self._select_reader_by_uri(gcs_object.name, content.reader)
            reader = cast(Reader, reader)

            # Fetch and load the content
            readable_content = BytesIO(gcs_object.download_as_bytes())

            # Read the content
            read_documents = await reader.async_read(readable_content, name=file_name)

            # Prepare and insert the content in the vector database
            self._prepare_documents_for_insert(read_documents, content_entry.id)
            await self._ahandle_vector_db_insert(content_entry, read_documents, upsert)

    def _load_from_gcs(
        self,
        content: Content,
        upsert: bool,
        skip_if_exists: bool,
        config: Optional[BaseStorageConfig] = None,
    ):
        """Load content from Google Cloud Storage (sync)."""
        try:
            from google.cloud import storage  # type: ignore  # noqa: F401
        except ImportError:
            raise ImportError(
                "The `google-cloud-storage` package is not installed. "
                "Please install it via `pip install google-cloud-storage`."
            )

        log_warning(
            "GCS content loading has limited features. "
            "Recursive folder traversal, rich metadata, and improved naming are coming in a future release."
        )

        remote_content: GCSContent = cast(GCSContent, content.remote_content)
        gcs_config = self._validate_gcs_config(content, config)

        # Get or create bucket
        bucket = remote_content.bucket
        if bucket is None and remote_content.bucket_name:
            client = self._get_gcs_client(gcs_config)
            bucket = client.bucket(remote_content.bucket_name)

        # Identify objects to read
        objects_to_read = []
        if remote_content.blob_name is not None:
            objects_to_read.append(bucket.blob(remote_content.blob_name))  # type: ignore
        elif remote_content.prefix is not None:
            objects_to_read.extend(bucket.list_blobs(prefix=remote_content.prefix))  # type: ignore
        else:
            objects_to_read.extend(bucket.list_blobs())  # type: ignore

        if objects_to_read:
            log_info(f"Processing {len(objects_to_read)} file(s) from GCS")

        bucket_name = remote_content.bucket_name or (bucket.name if bucket else "unknown")
        is_folder_upload = len(objects_to_read) > 1
        root_path = remote_content.prefix or ""

        for gcs_object in objects_to_read:
            blob_name = gcs_object.name
            file_name = blob_name.split("/")[-1]

            # Build metadata and virtual path using helpers
            virtual_path = self._build_gcs_virtual_path(bucket_name, blob_name)
            gcs_metadata = self._build_gcs_metadata(gcs_config, bucket_name, blob_name)
            merged_metadata: Dict[str, Any] = self._merge_metadata(gcs_metadata, content.metadata)

            # Compute content name using base helper
            content_name = self._compute_content_name(blob_name, file_name, content.name, root_path, is_folder_upload)

            # Create content entry
            content_entry = Content(
                name=content_name,
                description=content.description,
                path=virtual_path,
                status=ContentStatus.PROCESSING,
                metadata=merged_metadata,
                file_type="gcs",
            )
            content_entry.content_hash = self._build_content_hash(content_entry)
            content_entry.id = generate_id(content_entry.content_hash)

            self._insert_contents_db(content_entry)

            if self._should_skip(content_entry.content_hash, skip_if_exists):
                content_entry.status = ContentStatus.COMPLETED
                self._update_content(content_entry)
                continue

            # Select reader
            reader = self._select_reader_by_uri(gcs_object.name, content.reader)
            reader = cast(Reader, reader)

            # Fetch and load the content
            readable_content = BytesIO(gcs_object.download_as_bytes())

            # Read the content
            read_documents = reader.read(readable_content, name=file_name)

            # Prepare and insert the content in the vector database
            self._prepare_documents_for_insert(read_documents, content_entry.id)
            self._handle_vector_db_insert(content_entry, read_documents, upsert)



from typing import Optional, Union

from fastapi import HTTPException

from agno.agent import Agent, RemoteAgent
from agno.db.base import AsyncBaseDb, BaseDb
from agno.eval.accuracy import AccuracyEval
from agno.eval.agent_as_judge import AgentAsJudgeEval
from agno.eval.performance import PerformanceEval
from agno.eval.reliability import ReliabilityEval
from agno.models.base import Model
from agno.os.routers.evals.schemas import EvalRunInput, EvalSchema
from agno.team import RemoteTeam, Team


async def run_accuracy_eval(
    eval_run_input: EvalRunInput,
    db: Union[BaseDb, AsyncBaseDb],
    agent: Optional[Agent] = None,
    team: Optional[Team] = None,
    default_model: Optional[Model] = None,
) -> EvalSchema:
    """Run an Accuracy evaluation for the given agent or team"""
    if not eval_run_input.expected_output:
        raise HTTPException(status_code=400, detail="expected_output is required for accuracy evaluation")

    accuracy_eval = AccuracyEval(
        db=db,
        agent=agent,
        team=team,
        input=eval_run_input.input,
        expected_output=eval_run_input.expected_output,
        additional_guidelines=eval_run_input.additional_guidelines,
        additional_context=eval_run_input.additional_context,
        num_iterations=eval_run_input.num_iterations or 1,
        name=eval_run_input.name,
        model=default_model,
    )

    result = await accuracy_eval.arun(print_results=False, print_summary=False)
    if not result:
        raise HTTPException(status_code=500, detail="Failed to run accuracy evaluation")

    eval_run = EvalSchema.from_accuracy_eval(accuracy_eval=accuracy_eval, result=result)

    # Restore original model after eval
    if default_model is not None:
        if agent is not None:
            agent.model = default_model
        elif team is not None:
            team.model = default_model

    return eval_run


async def run_agent_as_judge_eval(
    eval_run_input: EvalRunInput,
    db: Union[BaseDb, AsyncBaseDb],
    agent: Optional[Union[Agent, RemoteAgent]] = None,
    team: Optional[Union[Team, RemoteTeam]] = None,
    default_model: Optional[Model] = None,
) -> EvalSchema:
    """Run an AgentAsJudge evaluation for the given agent or team"""
    if not eval_run_input.criteria:
        raise HTTPException(status_code=400, detail="criteria is required for agent-as-judge evaluation")

    # Run agent/team to get output
    if agent:
        agent_response = await agent.arun(eval_run_input.input, stream=False)  # type: ignore[misc]
        output = str(agent_response.content) if agent_response.content else ""
        agent_id = agent.id
        team_id = None
    elif team:
        team_response = await team.arun(eval_run_input.input, stream=False)  # type: ignore[misc]
        output = str(team_response.content) if team_response.content else ""
        agent_id = None
        team_id = team.id
    else:
        raise HTTPException(status_code=400, detail="Either agent_id or team_id must be provided")

    agent_as_judge_eval = AgentAsJudgeEval(
        db=db,
        criteria=eval_run_input.criteria,
        scoring_strategy=eval_run_input.scoring_strategy or "binary",
        threshold=eval_run_input.threshold or 7,
        additional_guidelines=eval_run_input.additional_guidelines,
        name=eval_run_input.name,
        model=default_model,
    )

    result = await agent_as_judge_eval.arun(
        input=eval_run_input.input, output=output, print_results=False, print_summary=False
    )
    if not result:
        raise HTTPException(status_code=500, detail="Failed to run agent as judge evaluation")

    # Use evaluator's model
    eval_model_id = agent_as_judge_eval.model.id if agent_as_judge_eval.model is not None else None
    eval_model_provider = agent_as_judge_eval.model.provider if agent_as_judge_eval.model is not None else None

    eval_run = EvalSchema.from_agent_as_judge_eval(
        agent_as_judge_eval=agent_as_judge_eval,
        result=result,
        agent_id=agent_id,
        team_id=team_id,
        model_id=eval_model_id,
        model_provider=eval_model_provider,
    )

    # Restore original model after eval
    if default_model is not None:
        if agent is not None and isinstance(agent, Agent):
            agent.model = default_model
        elif team is not None and isinstance(team, Team):
            team.model = default_model

    return eval_run


async def run_performance_eval(
    eval_run_input: EvalRunInput,
    db: Union[BaseDb, AsyncBaseDb],
    agent: Optional[Agent] = None,
    team: Optional[Team] = None,
    default_model: Optional[Model] = None,
) -> EvalSchema:
    """Run a performance evaluation for the given agent or team"""
    if agent:

        async def run_component():  # type: ignore
            return await agent.arun(eval_run_input.input, stream=False)  # type: ignore[misc]

        model_id = agent.model.id if agent and agent.model else None
        model_provider = agent.model.provider if agent and agent.model else None

    elif team:

        async def run_component():  # type: ignore
            return await team.arun(eval_run_input.input, stream=False)  # type: ignore[misc]

        model_id = team.model.id if team and team.model else None
        model_provider = team.model.provider if team and team.model else None

    performance_eval = PerformanceEval(
        db=db,
        name=eval_run_input.name,
        func=run_component,
        num_iterations=eval_run_input.num_iterations or 10,
        warmup_runs=eval_run_input.warmup_runs,
        agent_id=agent.id if agent else None,
        team_id=team.id if team else None,
        model_id=model_id,
        model_provider=model_provider,
    )

    result = await performance_eval.arun(print_results=False, print_summary=False)
    if not result:
        raise HTTPException(status_code=500, detail="Failed to run performance evaluation")

    eval_run = EvalSchema.from_performance_eval(
        performance_eval=performance_eval,
        result=result,
        agent_id=agent.id if agent else None,
        team_id=team.id if team else None,
        model_id=model_id,
        model_provider=model_provider,
    )

    # Restore original model after eval
    if default_model is not None:
        if agent is not None:
            agent.model = default_model
        elif team is not None:
            team.model = default_model

    return eval_run


async def run_reliability_eval(
    eval_run_input: EvalRunInput,
    db: Union[BaseDb, AsyncBaseDb],
    agent: Optional[Agent] = None,
    team: Optional[Team] = None,
    default_model: Optional[Model] = None,
) -> EvalSchema:
    """Run a reliability evaluation for the given agent or team"""
    if not eval_run_input.expected_tool_calls:
        raise HTTPException(status_code=400, detail="expected_tool_calls is required for reliability evaluations")

    if agent:
        agent_response = await agent.arun(eval_run_input.input, stream=False)  # type: ignore[misc]
        reliability_eval = ReliabilityEval(
            db=db,
            name=eval_run_input.name,
            agent_response=agent_response,
            expected_tool_calls=eval_run_input.expected_tool_calls,
        )
        model_id = agent.model.id if agent and agent.model else None
        model_provider = agent.model.provider if agent and agent.model else None

    elif team:
        team_response = await team.arun(eval_run_input.input, stream=False)  # type: ignore[misc]
        reliability_eval = ReliabilityEval(
            db=db,
            name=eval_run_input.name,
            team_response=team_response,
            expected_tool_calls=eval_run_input.expected_tool_calls,
        )
        model_id = team.model.id if team and team.model else None
        model_provider = team.model.provider if team and team.model else None

    result = await reliability_eval.arun(print_results=False)
    if not result:
        raise HTTPException(status_code=500, detail="Failed to run reliability evaluation")

    eval_run = EvalSchema.from_reliability_eval(
        reliability_eval=reliability_eval,
        result=result,
        agent_id=agent.id if agent else None,
        model_id=model_id,
        model_provider=model_provider,
    )

    # Restore original model after eval
    if default_model is not None:
        if agent is not None:
            agent.model = default_model
        elif team is not None:
            team.model = default_model

    return eval_run



from agno.models.cerebras.cerebras import Cerebras

try:
    from agno.models.cerebras.cerebras_openai import CerebrasOpenAI
except ImportError:

    class CerebrasOpenAI:  # type: ignore
        def __init__(self, *args, **kwargs):
            raise ImportError("`openai` not installed. Please install it via `pip install openai`")


__all__ = ["Cerebras", "CerebrasOpenAI"]



def quote_db_identifier(db_type: str, identifier: str) -> str:
    """Add the right quotes to the given identifier string (table name, schema name) based on db type.

    Args:
        db_type: The database type name (e.g., "PostgresDb", "MySQLDb", "SqliteDb")
        identifier: The identifier string to add quotes to

    Returns:
        The properly quoted identifier string
    """
    if db_type in ("MySQLDb", "AsyncMySQLDb", "SingleStoreDb"):
        escaped = identifier.replace("`", "``")
        return f"`{escaped}`"
    else:
        # Postgres, SQLite, and unknown types all use double-quote identifiers
        escaped = identifier.replace('"', '""')
        return f'"{escaped}"'



from dataclasses import dataclass, field
from os import getenv
from typing import Any, Dict, Optional

from agno.exceptions import ModelAuthenticationError
from agno.models.openai.like import OpenAILike


@dataclass
class InternLM(OpenAILike):
    """
    Class for interacting with the InternLM API.

    Attributes:
        id (str): The ID of the language model. Defaults to "internlm2.5-latest".
        name (str): The name of the model. Defaults to "InternLM".
        provider (str): The provider of the model. Defaults to "InternLM".
        api_key (Optional[str]): The API key for the InternLM API.
        base_url (Optional[str]): The base URL for the InternLM API.
    """

    id: str = "internlm2.5-latest"
    name: str = "InternLM"
    provider: str = "InternLM"

    api_key: Optional[str] = field(default_factory=lambda: getenv("INTERNLM_API_KEY"))
    base_url: Optional[str] = "https://internlm-chat.intern-ai.org.cn/puyu/api/v1/chat/completions"

    def _get_client_params(self) -> Dict[str, Any]:
        """
        Returns client parameters for API requests, checking for INTERNLM_API_KEY.

        Returns:
            Dict[str, Any]: A dictionary of client parameters for API requests.
        """
        if not self.api_key:
            self.api_key = getenv("INTERNLM_API_KEY")
            if not self.api_key:
                raise ModelAuthenticationError(
                    message="INTERNLM_API_KEY not set. Please set the INTERNLM_API_KEY environment variable.",
                    model_name=self.name,
                )
        return super()._get_client_params()



from dataclasses import dataclass
from os import getenv
from typing import Optional

from agno.knowledge.embedder.openai import OpenAIEmbedder


@dataclass
class NebiusEmbedder(OpenAIEmbedder):
    id: str = "BAAI/bge-en-icl"
    dimensions: int = 1024
    api_key: Optional[str] = getenv("NEBIUS_API_KEY")
    base_url: str = "https://api.tokenfactory.nebius.com/v1/"



from typing import Dict, List, Optional, Union

from fastapi.routing import APIRouter

from agno.agent import Agent, RemoteAgent
from agno.os.interfaces.base import BaseInterface
from agno.os.interfaces.telegram.router import (
    DEFAULT_ERROR_MESSAGE,
    DEFAULT_HELP_MESSAGE,
    DEFAULT_NEW_MESSAGE,
    DEFAULT_START_MESSAGE,
    attach_routes,
)
from agno.team import RemoteTeam, Team
from agno.workflow import RemoteWorkflow, Workflow

DEFAULT_BOT_COMMANDS: List[Dict[str, str]] = [
    {"command": "start", "description": "Start the bot"},
    {"command": "help", "description": "Show help"},
    {"command": "new", "description": "Start a new conversation"},
]


class Telegram(BaseInterface):
    type = "telegram"

    router: APIRouter

    def __init__(
        self,
        agent: Optional[Union[Agent, RemoteAgent]] = None,
        team: Optional[Union[Team, RemoteTeam]] = None,
        workflow: Optional[Union[Workflow, RemoteWorkflow]] = None,
        prefix: str = "/telegram",
        tags: Optional[List[str]] = None,
        token: Optional[str] = None,
        reply_to_mentions_only: bool = True,
        reply_to_bot_messages: bool = True,
        start_message: str = DEFAULT_START_MESSAGE,
        help_message: str = DEFAULT_HELP_MESSAGE,
        error_message: str = DEFAULT_ERROR_MESSAGE,
        streaming: bool = True,
        show_reasoning: bool = False,
        commands: Optional[List[Dict[str, str]]] = None,
        register_commands: bool = True,
        new_message: str = DEFAULT_NEW_MESSAGE,
    ):
        self.agent = agent
        self.team = team
        self.workflow = workflow
        self.prefix = prefix
        self.tags = tags or ["Telegram"]
        self.token = token
        self.reply_to_mentions_only = reply_to_mentions_only
        self.reply_to_bot_messages = reply_to_bot_messages
        self.start_message = start_message
        self.help_message = help_message
        self.error_message = error_message
        self.streaming = streaming
        self.show_reasoning = show_reasoning
        self.commands = commands if commands is not None else DEFAULT_BOT_COMMANDS
        self.register_commands = register_commands
        self.new_message = new_message

        if not (self.agent or self.team or self.workflow):
            raise ValueError("Telegram requires an agent, team, or workflow")

    def get_router(self) -> APIRouter:
        return attach_routes(
            router=APIRouter(prefix=self.prefix, tags=self.tags),  # type: ignore
            agent=self.agent,
            team=self.team,
            workflow=self.workflow,
            token=self.token,
            reply_to_mentions_only=self.reply_to_mentions_only,
            reply_to_bot_messages=self.reply_to_bot_messages,
            start_message=self.start_message,
            help_message=self.help_message,
            error_message=self.error_message,
            streaming=self.streaming,
            show_reasoning=self.show_reasoning,
            commands=self.commands,
            register_commands=self.register_commands,
            new_message=self.new_message,
        )
